import os

from typing import Dict

import e3.electrolyt.plan as plan
from e3.anod.context import AnodContext, SchedulingError
from e3.anod.error import AnodError
from e3.anod.loader import AnodSpecRepository
from e3.env import BaseEnv

import pytest


class TestContext:

    spec_dir = os.path.join(os.path.dirname(__file__), "context_data")

    def create_context(self, reject_duplicates: bool = True) -> AnodContext:
        """Create a spec repository and anod context.

        :param reject_duplicates: whether to reject duplicates in plan
        """

        def repo_conf(name: str) -> Dict[str, str]:
            return {"vcs": "git", "url": name, "branch": "master"}

        # Create a context for a x86-linux machine
        asr = AnodSpecRepository(self.spec_dir)
        asr.repos["spec1-git"] = repo_conf("spec1")
        asr.repos["spec8-git"] = repo_conf("spec8")
        asr.repos["spec2-git"] = repo_conf("spec2")
        asr.repos["a-git"] = repo_conf("a")
        env = BaseEnv()
        env.set_build("x86-linux", "rhes6", "mylinux")
        ac = AnodContext(asr, default_env=env, reject_duplicates=reject_duplicates)
        return ac

    def test_context_init(self):
        # Create a context using:
        # 1. the local default configuration
        # 2. forcing a x86-linux configuration
        asr = AnodSpecRepository(self.spec_dir)
        ac = AnodContext(asr)
        assert ac.default_env.build == BaseEnv().build
        assert ac.default_env.host == BaseEnv().host
        assert ac.default_env.target == BaseEnv().target
        self.create_context()

    def test_load(self):
        # Load a simple build specification that declares a single source
        ac = self.create_context()
        ac.load("spec1", env=ac.default_env, qualifier="", kind="build")

        # Load it a second time should use the cache data
        ac.load("spec1", env=None, qualifier="", kind="build")

        # One source should have been registered
        assert (
            len(ac.sources) == 1 and "spec1-src" in ac.sources
        ), "spec1-src source from spec1.anod has not been registered"

        # One spec instance should have been registered in the cache
        assert len(ac.cache) == 1, "caching of anod instances broken"

    def test_add_anod_action(self):
        # Load spec1 with build primitive
        ac = self.create_context()

        # the result should not be schedulable as there is no build
        # primitive defined in spec1
        with pytest.raises(SchedulingError):
            print(ac.add_anod_action("spec1", env=ac.default_env, primitive="build"))

    def test_add_anod_action_source(self):
        """Test source packaging."""
        ac = self.create_context()

        ac.add_anod_action("spec1", env=ac.default_env, primitive="source")
        result = ac.schedule(ac.always_create_source_resolver)
        assert len(result) == 5, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec1.source.spec1-src",
            "mylinux.x86-linux.spec1.source.sources",
            "checkout.spec1-git",
            "mylinux.x86-linux.spec1.upload_src.spec1-src",
        }

    def test_add_anod_action2(self):
        # Simple spec with sources associated to the build primitive
        ac = self.create_context()
        ac.add_anod_action("spec2", env=ac.default_env, primitive="build")
        assert len(ac.tree) == 9, ac.tree.as_dot()

        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 5, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec2.build",
            "source_get.spec2-src",
            "mylinux.x86-linux.spec2.source_install.spec2-src",
            "download.spec2-src",
        }

    def test_add_anod_action2_force_install(self):
        """Check that forcing an install with no package is rejected."""
        ac = self.create_context()
        try:
            ac.add_anod_action(
                "spec2",
                env=ac.default_env,
                primitive="install",
                plan_args={},
                plan_line="install_plan.txt:2",
            )
        except SchedulingError as err:
            assert (
                "error in plan at install_plan.txt:2: install should "
                "be replaced by build" in str(err)
            )

    def test_add_anod_action2_no_source_resolver(self):
        def no_resolver(action, decision):
            return AnodContext.decision_error(action, decision)

        ac = self.create_context()
        ac.add_anod_action("spec2", env=ac.default_env, primitive="build")
        assert len(ac.tree) == 9, ac.tree.as_dot()

        with pytest.raises(SchedulingError) as err:
            ac.schedule(no_resolver)
        assert (
            "This plan resolver cannot decide whether what to do"
            " for resolving source_get.spec2-src." in str(err)
        )

    def test_add_anod_action3(self):
        # Simple spec with both install and build primitive and a package
        # declared
        ac = self.create_context()
        ac.add_anod_action("spec3", env=ac.default_env, primitive="build")
        assert len(ac.tree) == 6, ac.tree.as_dot()
        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 4, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec3.build",
            "mylinux.x86-linux.spec3.install",
            "mylinux.x86-linux.spec3.upload_bin",
        }

    def test_add_anod_action4(self):
        # Simple spec with:
        #   install primitive, package, component
        #   build primitive
        ac = self.create_context()
        ac.add_anod_action("spec4", env=ac.default_env, primitive="build")
        assert len(ac.tree) == 6, ac.tree.as_dot()
        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 4, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec4.build",
            "mylinux.x86-linux.spec4.install",
            "mylinux.x86-linux.spec4.upload_bin",
        }

    def test_add_anod_action4_2(self):
        # Same previous example but calling install primitive instead of build
        ac = self.create_context()
        ac.add_anod_action("spec4", env=ac.default_env, primitive="install")
        assert len(ac.tree) == 5, ac.tree.as_dot()
        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 3, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec4.download_bin",
            "mylinux.x86-linux.spec4.install",
        }

    def test_add_anod_action4_3(self):
        # Same as previous example but calling test primitive
        ac = self.create_context()
        ac.add_anod_action("spec4", env=ac.default_env, primitive="test")
        assert len(ac.tree) == 2, ac.tree.as_dot()
        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 2, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec4.test",
        }

    def test_add_anod_action5(self):
        # Case in which a source component should be uploaded (i.e: no binary
        # package declared)
        ac = self.create_context()
        ac.add_anod_action("spec5", env=ac.default_env, primitive="build")
        assert len(ac.tree) == 3, ac.tree.as_dot()
        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 3, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec5.build",
            "mylinux.x86-linux.spec5.upload_bin",
        }

    def test_add_anod_action6(self):
        # Calling install on a spec without install primitive result in a build
        # ??? should we allow that ???
        ac = self.create_context()
        ac.add_anod_action("spec6", env=ac.default_env, primitive="install")
        assert len(ac.tree) == 2, ac.tree.as_dot()
        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 2, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec6.build",
        }

    def test_add_anod_action6_2(self):
        # Same as previous example. Just ensure that if the spec is called
        # twice with different qualifiers that have no effect on build space
        # name then the result is only one install. (and thus qualifier value
        # for that node won't be deterministic.
        # ??? Should we raise issues on such cases ???
        ac = self.create_context()
        ac.add_anod_action("spec6", env=ac.default_env, primitive="install")
        ac.add_anod_action(
            "spec6", env=ac.default_env, primitive="install", qualifier="myqualif"
        )
        assert len(ac.tree) == 2, ac.tree.as_dot()
        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 2, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec6.build",
        }

    def test_add_anod_action7(self):
        # Ensure that build_deps = None is accepted
        ac = self.create_context()
        ac.add_anod_action("spec7", env=ac.default_env, primitive="build")
        result = ac.schedule(ac.always_download_source_resolver)
        assert len(result) == 2, result.as_dot()
        assert set(result.vertex_data.keys()) == {
            "root",
            "mylinux.x86-linux.spec7.build",
        }

    def test_add_anod_action8(self):
        """Simple spec with source that does not exist."""
        ac = self.create_context()
        with pytest.raises(AnodError):
            ac.add_anod_action("spec8", env=ac.default_env, primitive="build")

    def test_add_anod_action9(self):
        """Test source dependency."""
        ac = self.create_context()
        ac.add_anod_action("spec9", env=ac.default_env, primitive="build")
        result = ac.schedule(ac.always_download_source_resolver)
        assert "download.spec2-src" in list(result.vertex_data.keys())

    def test_add_anod_action10(self):
        """Verify that requiring both build and install fails."""
        ac = self.create_context()
        ac.add_anod_action("spec3", env=ac.default_env, primitive="install")
        ac.add_anod_action("spec3", env=ac.default_env, primitive="build")

        with pytest.raises(SchedulingError):
            ac.schedule(ac.always_download_source_resolver)

    def test_add_anod_action11_build_tree_dep(self):
        """Check build dependencies."""
        ac = self.create_context()
        ac.add_anod_action(
            "spec10",
            env=ac.default_env,
            qualifier="build_tree",
            primitive="build",
            plan_line="myplan:1",
        )

        # we have a dep on spec3 build_pkg, we require an explicit call to
        # build
        with pytest.raises(SchedulingError) as err:
            ac.schedule(ac.always_download_source_resolver)

        assert "has a build_tree dependency on spec3" in str(err)
        assert 'anod_build("spec3", qualifier="foo"' ', build="x86-linux")' in str(err)

        ac.add_anod_action(
            "spec3",
            env=ac.default_env,
            qualifier="foo",
            primitive="install",
            plan_line="myplan:2",
        )
        with pytest.raises(SchedulingError) as err:
            ac.schedule(ac.always_download_source_resolver)

        assert "explicit DownloadBinary decision made by myplan:2" in str(err)

    def test_add_anod_action11_build_tree_dep_with_env(self):
        """Check build dependencies."""
        ac = self.create_context()
        ac.add_anod_action(
            "spec10",
            env=ac.default_env,
            qualifier="build_tree_with_env",
            primitive="build",
            plan_line="myplan:1",
        )

        # we have a dep on spec3 build_pkg, we require an explicit call to
        # build
        with pytest.raises(SchedulingError) as err:
            ac.schedule(ac.always_download_source_resolver)

        assert "has a build_tree dependency on spec3" in str(err.value)
        assert (
            'anod_build("spec3", qualifier="foo",'
            ' build="x86_64-linux", host="x86-linux",'
            ' target="arm-elf")' in str(err.value)
        )

    def test_add_anod_action11_install_dep(self):
        """Check build dependencies."""
        ac = self.create_context()
        ac.add_anod_action(
            "spec10", env=ac.default_env, primitive="build", plan_line="myplan:1"
        )

        # we have a dep on spec3 build_pkg, we require an explicit call to
        # build
        with pytest.raises(SchedulingError) as err:
            ac.schedule(ac.always_download_source_resolver)

        assert (
            'Please either add anod_build("spec3", build="x86-linux") '
            'or anod_install("spec3", build="x86-linux")'
            " in the plan" in str(err)
        )

        ac.add_anod_action(
            "spec3", env=ac.default_env, primitive="install", plan_line="myplan:2"
        )

        # This call should not raise an exception
        ac.schedule(ac.always_download_source_resolver)

    def test_add_anod_action12(self):
        """Check handling of duplicated source package."""
        ac = self.create_context()
        ac.add_anod_action("spec12", env=ac.default_env, primitive="build")
        result = ac.schedule(ac.always_download_source_resolver)
        keys = set(result.vertex_data.keys())
        assert "download.spec1-src" in keys
        assert "download.unmanaged-src" in keys

    def test_add_anod_action13(self):
        """Check handling of install without build."""
        ac = self.create_context()
        ac.add_anod_action("spec13", env=ac.default_env, primitive="install")
        result = ac.schedule(ac.always_download_source_resolver)
        keys = set(result.vertex_data.keys())
        assert len(keys) == 3, keys
        assert "mylinux.x86-linux.spec13.download_bin" in keys
        assert "mylinux.x86-linux.spec13.install" in keys

    def test_add_anod_action_unmanaged_source(self):
        """Check no source creation for thirdparties."""
        ac = self.create_context()
        ac.add_anod_action(
            "spec-unmanaged-source", env=ac.default_env, primitive="source"
        )
        result = ac.schedule(ac.always_download_source_resolver)
        keys = set(result.vertex_data.keys())
        assert len(keys) == 2, keys
        assert "mylinux.x86-linux.spec-unmanaged-source.source.wheel.whl" not in keys

    def test_add_anod_action_managed_source(self):
        """Check no source creation for thirdparties."""
        ac = self.create_context()
        ac.add_anod_action(
            "spec-managed-source", env=ac.default_env, primitive="source"
        )
        result = ac.schedule(ac.always_download_source_resolver)
        keys = set(result.vertex_data.keys())
        assert len(keys) == 5, keys
        assert "checkout.a-git" in keys
        assert "mylinux.x86-linux.spec-managed-source.source.a-src" in keys

    def test_dag_2_plan(self):
        """Check that we can extract values from plan in final dag.

        Some paramaters passed in the plan are lost in the final
        scheduled dag, when plan lines are transformed into
        anod actions. It is possible to retrieve them by looking
        at the tags.
        """
        # Create a new plan context
        ac = self.create_context()
        current_env = BaseEnv()
        cm = plan.PlanContext(server=current_env)

        # Declare available actions and their signature
        def anod_action(
            module,
            build=None,
            default_build=False,
            host=None,
            target=None,
            board=None,
            weathers=None,
            product_version=None,
            when_missing=None,
            manual_action=False,
            qualifier=None,
            jobs=None,
            releases=None,
            process_suffix=None,
            update_vcs=False,
            recursive=None,
            query_range=None,
            force_repackage=False,
        ):
            pass

        for a in ("anod_build", "anod_install", "anod_source", "anod_test"):
            cm.register_action(a, anod_action)

        # Create a simple plan
        content = [
            "def myserver():",
            '    anod_build("spec12", weathers="foo")',
            '    anod_build("spec10", weathers="foo")',
            '    anod_build("spec11", weathers="bar")',
        ]
        with open("plan.txt", "w") as f:
            f.write("\n".join(content))
        myplan = plan.Plan({}, plan_ext=".txt")
        myplan.load("plan.txt")

        # Execute the plan and create anod actions
        for action in cm.execute(myplan, "myserver"):
            primitive = action.action.replace("anod_", "", 1)
            ac.add_anod_action(
                name=action.module,
                env=current_env if action.default_build else action,
                primitive=primitive,
                qualifier=action.qualifier,
                plan_line=action.plan_line,
                plan_args=action.plan_args,
            )

        # Create a reverse tag to have a working get_context
        # when looking for parameters such as weathers we want to
        # get the plan line that has triggered the action, e.g.
        # for spec3.build that has been triggered by spec10.build
        # we want to propagate the weathers set in the line
        #     anod_build("spec10", weathers="foo")
        # in the Build action for spec3
        reverse_dag = ac.tree.reverse_graph()

        for uid, action in ac.tree:
            if uid.endswith("spec12.build"):
                assert ac.tree.get_tag(uid)
                cdist, cuid, ctag = reverse_dag.get_context(uid)[0]
                assert cuid == uid
                assert ctag["plan_args"]["weathers"] == "foo"
                assert ctag["plan_line"] == "plan.txt:2"
            elif uid.endswith("spec3.build"):
                assert not ac.tree.get_tag(uid)
                cdist, cuid, ctag = reverse_dag.get_context(uid)[0]
                assert cuid != uid
                assert cuid.endswith("spec10.build")
                assert ctag["plan_args"]["weathers"] == "foo"
                assert ctag["plan_line"] == "plan.txt:3"
            elif uid.endswith("spec11.build"):
                assert ac.tree.get_tag(uid), ac.tree.tags
                cdist, cuid, ctag = reverse_dag.get_context(uid)[0]
                assert cuid == uid
                assert ctag["plan_args"]["weathers"] == "bar"
                assert ctag["plan_line"] == "plan.txt:4"

                # Also verify that the instance deps is properly loaded
                assert set(action.anod_instance.deps.keys()) == {"spec1"}
                assert action.anod_instance.deps["spec1"].__class__.__name__ == "Spec1"

        # Also test that we are still able to extract the values
        # after having scheduled the action graph.

        # Create an explict build action to make sure that the plan can be
        # scheduled
        ac.add_anod_action(
            name="spec3",
            env=current_env,
            primitive="build",
            plan_line="plan.txt:5",
            plan_args={"weathers": "my_spec3_weather"},
        )

        sched_dag = ac.schedule(ac.always_download_source_resolver)
        sched_rev = sched_dag.reverse_graph()

        for uid, action in sched_dag:
            if uid.endswith("spec12.build"):
                assert sched_dag.get_tag(uid)

                # Also verify that the instance deps is properly loaded
                assert set(action.anod_instance.deps.keys()) == {"spec1", "spec11"}
                assert (
                    action.anod_instance.deps["spec11"].__class__.__name__ == "Spec11"
                )
                assert action.anod_instance.deps["spec1"].__class__.__name__ == "Spec1"

            elif uid.endswith("spec3.build"):
                assert sched_dag.get_tag(uid)
                assert (
                    sched_rev.get_context(uid)[0][2]["plan_args"]["weathers"]
                    == "my_spec3_weather"
                )

    def test_dag_2_plan_sources(self):
        """Check that we can extract values from plan in final dag.

        Use a scheduler to always create source and ask for a source
        package creation.
        """
        # Create a new plan context
        ac = self.create_context()
        current_env = BaseEnv()
        cm = plan.PlanContext(server=current_env)

        # Declare available actions and their signature
        def anod_action(
            module,
            build=None,
            default_build=False,
            host=None,
            target=None,
            board=None,
            weathers=None,
            product_version=None,
            when_missing=None,
            manual_action=False,
            qualifier=None,
            jobs=None,
            releases=None,
            process_suffix=None,
            update_vcs=False,
            recursive=None,
            query_range=None,
            force_repackage=False,
        ):
            pass

        cm.register_action("anod_source", anod_action)

        # Create a simple plan
        content = ["def myserver():", '    anod_source("spec1", weathers="foo")']
        with open("plan.txt", "w") as f:
            f.write("\n".join(content))
        myplan = plan.Plan({})
        myplan.load("plan.txt")

        # Execute the plan and create anod actions
        for action in cm.execute(myplan, "myserver"):
            primitive = action.action.replace("anod_", "", 1)
            ac.add_anod_action(
                name=action.module,
                env=current_env if action.default_build else action,
                primitive=primitive,
                qualifier=action.qualifier,
                plan_line=action.plan_line,
                plan_args=action.plan_args,
            )

        for uid, _ in ac.tree:
            if uid.endswith("sources"):
                assert ac.tree.get_tag(uid)
            elif uid.endswith(".source.spec1-src"):
                assert ac.tree.get_tag(uid)
                assert (
                    ac.tree.get_context(
                        vertex_id=uid, reverse_order=True, max_distance=1
                    )[0][2]["plan_args"]["weathers"]
                    == "foo"
                )

    @pytest.mark.parametrize("reject_duplicates", [True, False])
    def test_duplicated_lines(self, reject_duplicates):
        """Check that duplicated lines in plan are properly rejected."""
        ac = self.create_context(reject_duplicates=reject_duplicates)
        current_env = BaseEnv()
        cm = plan.PlanContext(server=current_env)

        # Declare available actions and their signature
        def anod_action(
            module,
            build=None,
            default_build=False,
            host=None,
            target=None,
            board=None,
            weathers=None,
            product_version=None,
            when_missing=None,
            manual_action=False,
            qualifier=None,
            jobs=None,
            releases=None,
            process_suffix=None,
            update_vcs=False,
            recursive=None,
            query_range=None,
            force_repackage=False,
        ):
            pass

        cm.register_action("anod_build", anod_action)
        # Create a simple plan
        content = [
            "def myserver():",
            '    anod_build("spec3", weathers="A")',
            '    anod_build("spec3", weathers="B")',
        ]
        with open("plan.plan", "w") as f:
            f.write("\n".join(content))
        myplan = plan.Plan({})
        myplan.load("plan.plan")

        if not reject_duplicates:
            # Execute the plan and create anod actions
            # Execute the plan and create anod actions
            for action in cm.execute(myplan, "myserver"):
                primitive = action.action.replace("anod_", "", 1)
                ac.add_anod_action(
                    name=action.module,
                    env=current_env if action.default_build else action,
                    primitive=primitive,
                    qualifier=action.qualifier,
                    plan_line=action.plan_line,
                    plan_args=action.plan_args,
                )

            for uid, _ in ac.tree:
                if uid.endswith("build"):
                    assert ac.tree.get_tag(uid)["plan_args"]["weathers"] == "B"
        else:

            with pytest.raises(SchedulingError):
                for action in cm.execute(myplan, "myserver"):
                    primitive = action.action.replace("anod_", "", 1)
                    ac.add_anod_action(
                        name=action.module,
                        env=current_env if action.default_build else action,
                        primitive=primitive,
                        qualifier=action.qualifier,
                        plan_line=action.plan_line,
                        plan_args=action.plan_args,
                    )

    def test_plan_call_args(self):
        """Retrieve call args values."""
        current_env = BaseEnv()
        cm = plan.PlanContext(server=current_env)

        # Declare available actions and their signature
        def plan_action(platform):
            pass

        cm.register_action("plan_action", plan_action)
        # Create a simple plan
        content = ["def myserver():", '    plan_action("any")']
        with open("plan.txt", "w") as f:
            f.write("\n".join(content))
        myplan = plan.Plan({})
        myplan.load("plan.txt")

        for action in cm.execute(myplan, "myserver"):
            assert action.plan_call_args == {"platform": "any"}
            assert action.plan_args["platform"] == BaseEnv().platform

    def test_add_anod_action_duplicate_dep(self):
        """Verify that duplicate dep with same local_name are rejected."""
        ac = self.create_context()
        with pytest.raises(AnodError) as err:
            ac.add_anod_action("duplicate_dep", env=ac.default_env, primitive="build")

        assert (
            "The spec duplicate_dep has two dependencies with the same "
            "local_name attribute (spec3)" in str(err)
        )



import email
import json
import os

import e3.archive
import e3.event
import e3.fs
import e3.os.fs

import mock


def test_smtp_event():

    e3.event.add_handler(
        "smtp",
        subject="test subject",
        from_addr="e3@example.net",
        to_addr="info@example.net,info@example.com",
        smtp_servers=["smtp.localhost"],
    )

    e3.event.add_handler("file", log_dir="./log")

    e3.fs.mkdir("log")
    e3.fs.mkdir("pkg")
    e3.archive.create_archive(
        filename="pkg.tar.gz",
        from_dir=os.path.join(os.getcwd(), "pkg"),
        dest=os.getcwd(),
    )

    e3.os.fs.touch("unknown")

    with e3.event.Event(name="event test") as e:
        e.attach_file("pkg.tar.gz", name="pkg.tar.gz")
        e.attach_file(__file__, name="test.py")
        e.attach_file("unknown", name="unknown")
        e.attr = "a test attr"

    with mock.patch("smtplib.SMTP_SSL") as mock_smtp:
        e3.event.send_event(e)

    smtp_mock = mock_smtp.return_value
    assert smtp_mock.sendmail.called
    assert smtp_mock.sendmail.call_count == 1
    msg_as_string = smtp_mock.sendmail.call_args[0][2]

    event_mail = email.message_from_string(msg_as_string)
    assert event_mail["Subject"] == "test subject"

    evdata = None
    attachment_content = None
    for part in event_mail.walk():
        part_type = part.get_content_maintype()
        subpart_type = part.get_content_type().split("/")[1]
        if part_type == "multipart":
            pass
        elif part_type == "application" and subpart_type == "json":
            part_content = part.get_payload(decode=True)
            evdata = json.loads(part_content.decode("utf-8"))
        elif part.get_filename() == "test.py":
            attachment_content = part.get_payload(decode=True)

    assert evdata is not None, "no json part found"
    assert evdata["name"] == "event test"
    assert evdata["attr"] == "a test attr"
    assert attachment_content is not None

    # attachment content contains this file content, including
    # the previous line
    assert b"assert attachment_content is not None" in attachment_content

    # Test sending an event in two steps: save in temporary files and then
    # reread the file and send the event.
    filename = e.dump(event_dir=".")
    assert os.path.isfile(filename)

    with mock.patch("smtplib.SMTP_SSL") as mock_smtp:
        e3.event.send_event_from_file(filename)

    smtp_mock = mock_smtp.return_value
    assert smtp_mock.sendmail.called
    assert smtp_mock.sendmail.call_count == 1
    msg_as_string = smtp_mock.sendmail.call_args[0][2]

    event_mail = email.message_from_string(msg_as_string)
    assert event_mail["Subject"] == "test subject"


def test_smtp_servers_as_str():
    """Test when smtp server setting is a string."""
    e3.event.add_handler(
        "smtp",
        subject="test subject",
        from_addr="e3@example.net",
        to_addr="info@example.net,info@example.com",
        smtp_servers="smtp.localhost",
    )
    assert isinstance(e3.event.default_manager.handlers["smtp"].smtp_servers, list)



#
# e3-core documentation build configuration file, created by
# sphinx-quickstart on Fri May 19 23:56:17 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
# type: ignore

import sphinx_rtd_theme


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["autoapi.extension"]

autoapi_type = "python"
autoapi_dirs = ["../../src/e3"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
autoapi_template_dir = "source/autoapi_templates"

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "e3-core"
copyright = "2017, AdaCore"
author = "AdaCore"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "21.0"
# The full version, including alpha/beta/rc tags.
release = "21.0"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "e3-coredoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "e3-core.tex", "e3-core Documentation", "AdaCore", "manual")
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "e3-core", "e3-core Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "e3-core",
        "e3-core Documentation",
        author,
        "e3-core",
        "One line description of project.",
        "Miscellaneous",
    )
]



import os
from collections import OrderedDict

import e3.log
import e3.yaml
import yaml
import yaml.constructor

import pytest

try:
    from io import StringIO
except ImportError:
    from io import StringIO


@pytest.mark.parametrize(
    "config,expected",
    [
        (
            {"param1": "full", "param2": "short"},
            {"y": 0, "c": "", "value2": "", "value1": 10},
        ),
        (
            {"param1": "short", "param2": "full"},
            {
                "a": 42,
                "c": ["1", "default", "2"],
                "b": 5,
                "value3": 30,
                "value2": "['default']",
                "value1": 10,
            },
        ),
        (
            {"param1": "short", "param2": "dict"},
            {
                "a": 9,
                "c": {"default": "ok", "default2": "nok"},
                "b": 5,
                "value2": "{'default': 'ok'}",
                "value1": 10,
            },
        ),
    ],
)
def test_case_parser(config, expected):
    """Test yaml CaseParser."""
    yaml_case_content = """
case_param1:
    full:
        case_param2:
            full: {'a': 2, 'b': 1, 'c': 'content'}
            short: {'y': 0, 'c': ''}
    short:
        case_param2:
            full: {'a': 9, 'b': 5, 'c': ['default']}
            short: {'y': 3, 'c': ''}
            dict: {'a': 9, 'b': 5, 'c': {'default': 'ok'}}

value1: 10
value2: '%(c)s'
case_param2:
    'f.*l': {'value3': 30, 'a': 42, '+c': ['2'], 'c+': ['1']}
    'dict': {'+c': {'default2': 'nok'}}
"""
    d = yaml.load(StringIO(yaml_case_content), e3.yaml.OrderedDictYAMLLoader)
    parse_it = e3.yaml.CaseParser(config).parse(d)
    assert parse_it == expected

    with open("tmp", "w") as f:
        f.write(yaml_case_content)

    parse_it2 = e3.yaml.load_with_config("tmp", config)
    assert parse_it2 == expected


def test_case_parser_object():
    """CaseParser supports any objects."""
    yaml_case_content = """
case_v:
    .*:
        dt: !!python/object/apply:time.gmtime []

result: '%(dt)s'
    """
    d = yaml.load(StringIO(yaml_case_content), e3.yaml.OrderedDictYAMLLoader)
    parse_it = e3.yaml.CaseParser({"v": "true"}).parse(d)
    assert "time.struct_time" in parse_it["result"]


def test_case_parser_err():
    """Test CaseParser error handling."""
    yaml_case_content = "result: '%(dt)s'"
    d = yaml.load(StringIO(yaml_case_content), e3.yaml.OrderedDictYAMLLoader)
    parse_it = e3.yaml.CaseParser({}).parse(d)
    assert parse_it["result"] == "%(dt)s"


def test_include():
    """Test yaml !include."""
    with open("1.yaml", "w") as f:
        f.write("b: !include 2.yaml\n")
        f.write("c: !include %s\n" % os.path.join(os.getcwd(), "2.yaml"))

    with open("2.yaml", "w") as f:
        f.write("a: 4\n")

    d = e3.yaml.load_ordered("1.yaml")
    assert d == OrderedDict(
        [("b", OrderedDict([("a", 4)])), ("c", OrderedDict([("a", 4)]))]
    )

    with pytest.raises(IOError) as err:
        yaml.load("b: !include foo.yaml\n", e3.yaml.OrderedDictYAMLLoader)
    assert "No such file or directory" in str(err.value)
    assert "foo.yaml" in str(err.value)


def test_duplicatekey():
    """Duplicated key should be rejected by load_ordered."""
    with open("dup.yaml", "w") as f:
        f.write("b: 2\nb: 9")

    with pytest.raises(yaml.constructor.ConstructorError) as err:
        e3.yaml.load_ordered("dup.yaml")
    assert "found duplicate key (b)" in str(err.value)


def test_yaml_err():
    """Test load_ordered error handling."""
    with open("err.yaml", "w") as f:
        f.write("[1]: 2\n")

    with pytest.raises(yaml.constructor.ConstructorError) as err:
        e3.yaml.load_ordered("err.yaml")
    assert "found unacceptable key" in str(err.value)

    with open("err2.yaml", "w") as f:
        f.write("--- !!map [not, a, map]\n")

    with pytest.raises(yaml.constructor.ConstructorError) as err:
        e3.yaml.load_ordered("err2.yaml")
    assert "expected a mapping node" in str(err.value)


def test_load_with_config_err():
    """Test load_with_config error handling."""
    with pytest.raises(e3.yaml.YamlError) as err:
        e3.yaml.load_with_config("/does/not/exist", {})
    assert "cannot read" in str(err.value)

    with open("err.yaml", "w") as f:
        f.write('"o" "o" "o"')

    with pytest.raises(e3.yaml.YamlError) as err:
        e3.yaml.load_with_config("err.yaml", {})
    assert "invalid yaml" in str(err.value)


def test_load_with_regexp():
    """Test load_with_regexp."""
    with open("regexp1.yaml", "w") as f:
        f.write(
            "key1: [['ppc-vx6-.*', 'qemu-toto', 'FALSE'],"
            " ['ppc-vx6-.*', 'qemu.*', '%(data)s']]\n"
        )
        f.write(
            "key2: [['ppc-vx6-linux', '', ['TRUE', '%(data)s']],"
            " ['ppc-vx6.*', 'qemu.*', ['FALSE']]]\n"
        )
    selectors = ["ppc-vx6-linux", "qemu"]
    data = {"data": "TRUE"}

    result = e3.yaml.load_with_regexp_table(
        filename="regexp1.yaml", selectors=selectors, data=data
    )

    assert len(result) == 2
    assert result["key1"] == "TRUE"
    assert result["key2"][0] == "TRUE"
    assert result["key2"][1] == "TRUE"



from __future__ import annotations

import collections.abc
from typing import TYPE_CHECKING

import e3.anod.error
from e3.env import BaseEnv

if TYPE_CHECKING:
    from typing import Any, Hashable, Optional
    from e3.anod.spec import Anod


class BuildVar:
    """Declare a dependency between an Anod spec and a variable."""

    def __init__(self, name: str, value: Hashable):
        """Initialize a BuildVar object.

        :param name: name of the variable
        :param value: variable value
        """
        assert isinstance(value, collections.abc.Hashable)
        self.name = name
        self.value = value
        self.kind = "var"

    def __str__(self) -> str:
        return f"{self.name}={self.value}"


class Dependency:
    def __init__(
        self,
        name: str,
        product_version: Optional[str] = None,
        host: Optional[str] = None,
        target: Optional[str] = None,
        build: Optional[str] = None,
        qualifier: Optional[str] = None,
        local_name: Optional[str] = None,
        require: str = "build_tree",
        track: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize a Dependency object.

        :param name: basename of the Anod spec file (without .anod extension)
        :param product_version: product version to force when loading the spec
        :param host: can be either 'target' or 'build'. If set to 'target' then
            it means that the dependency will be loaded with host set to
            the current module target information. If set to 'build' then
            similar mechanism is used with the current module 'build'
            platform.
        :param target: can be either 'host' or 'build'
        :param build: build platform (if not the current build platform), the
            special value "default" change the build platform to the default
            build platform. Note that on windows, the default is always
            x86-windows.
        :param qualifier: qualifier to set when loading the spec
        :param local_name: if not None, this name will be the dependency key
            in deps and makedeps dictionaries. It allows importing twice
            the same anod module with different qualifers or platforms
        :param require: can be 'build_tree' (to force a local build),
            'installation' or 'source_pkg'
        :param track: if True, track all source packages metadata and include
            them in the local metadata.
        :param kwargs: other parameters valid in some API that we ignore now
        """
        del kwargs
        self.name = name
        self.product_version = product_version
        self.host = host
        self.target = target
        self.build = build
        self.qualifier = qualifier
        self.local_name = local_name if local_name is not None else name
        if require not in ("build_tree", "installation", "source_pkg"):
            raise e3.anod.error.SpecError(
                "require should be build_tree, installation or source_pkg"
                " not %s." % require
            )
        self.kind = {
            "build_tree": "build",
            "installation": "install",
            "source_pkg": "source",
        }[require]
        self.track = track

    def env(self, parent: Anod, default_env: BaseEnv) -> BaseEnv:
        """Retrieve env for the dependency.

        :param parent: Anod instance in which the dep was declared
        :param default_env: default env for the current context
        :return: env object that should be used by the dependency
        """
        # Get the current environment associated with the Anod instance
        # and adjust it based on dependency parameters
        dep_env = BaseEnv(parent.env.build, parent.env.host, parent.env.target)

        # For simulation purposes we sometimes load specs as if it was
        # loaded on a non local machine thus 'default' does not correspond
        # to the default build platform of the local machine.
        if self.build == "default":
            build = default_env.build.platform
        else:
            build = self.build

        if self.host == "default":
            host = default_env.build.platform
        else:
            host = self.host

        if self.target == "default":
            target = default_env.build.platform
        else:
            target = self.target

        dep_env.set_env(build, host, target)
        return dep_env



import datetime
import sys
import json
import dateutil.parser
import e3.log
import e3.os.process
from logging import LogRecord


def test_log():
    p = e3.os.process.Run(
        [
            sys.executable,
            "-c",
            "\n".join(
                (
                    "import e3.log",
                    'e3.log.activate(filename="log.txt")',
                    'l = e3.log.getLogger("test_log")',
                    'l.debug("this is a log record")',
                )
            ),
        ]
    )
    assert p.status == 0

    with open("log.txt") as f:
        line = f.readline()
        # Get datetime in the log
        log_datetime, _, _ = line.partition(": ")

        # Parse it and verify that it is it in GMT
        assert (
            datetime.datetime.utcnow() - dateutil.parser.parse(log_datetime)
        ).seconds < 10


def test_json_log():
    """test logger method wrappers and json logs."""
    p = e3.os.process.Run(
        [
            sys.executable,
            "-c",
            "\n".join(
                (
                    "import e3.log",
                    'e3.log.activate(filename="log.json",json_format=True)',
                    'l = e3.log.getLogger("test_log")',
                    'l.debug("this is a log record")',
                    'l.info("Log record with extra param", anod_uui=1212)',
                    'l.warning("Log record with extra param", anod_uui=1212)',
                    'l.debug("Log record with extra param", anod_uui=1212)',
                    'l.error("Log record with extra param", anod_uui=1212)',
                    'l.critical("Log record with extra param", anod_uui=1212)',
                )
            ),
        ]
    )

    assert p.status == 0

    with open("log.json") as f:
        lines = f.readlines()

    record = json.loads(lines[0])
    # verify if we get default json fields
    assert len(record.keys()) == 5

    # verify presence of custom size field
    for line in lines[1:]:
        record = json.loads(line)
        assert "anod_uui" in record
        assert len(record.keys()) == 6


def test_json_log_compat():
    """make sure that code do not crash if json is not activated."""
    p = e3.os.process.Run(
        [
            sys.executable,
            "-c",
            "\n".join(
                (
                    "import e3.log",
                    'e3.log.activate(filename="log.json")',
                    'l = e3.log.getLogger("test_log")',
                    'l.debug("this is a log record")',
                    'l.debug("extra key",anod_uui=121222)',
                )
            ),
        ]
    )

    assert p.status == 0


def test_json_log_exception():
    """test exception attribute support."""
    p = e3.os.process.Run(
        [
            sys.executable,
            "-c",
            "\n".join(
                (
                    "import e3.log",
                    'e3.log.activate(filename="log.json", json_format=True)',
                    'l = e3.log.getLogger("test_log")',
                    "try:",
                    "    1/0",
                    "except Exception as e:",
                    "    l.exception(e,exc_info=True)",
                )
            ),
        ]
    )

    assert p.status == 0

    with open("log.json") as f:
        lines = f.readlines()

    record = json.loads(lines[0])
    assert "exc_text" in record


def test_json_context():
    """test context attribute when console_logs is used."""
    p = e3.os.process.Run(
        [
            sys.executable,
            "-c",
            "\n".join(
                (
                    "import e3.log",
                    "e3.log.console_logs='test_json'",
                    'e3.log.activate(filename="log.json",json_format=True)',
                    'l = e3.log.getLogger("test_log")',
                    'l.debug("message")',
                )
            ),
        ]
    )

    assert p.status == 0

    with open("log.json") as f:
        lines = f.readlines()

    record = json.loads(lines[0])
    assert "context" in record
    assert record["context"] == "test_json"


def test_json_formatter():
    """test json formatter."""
    formatter = e3.log.JSONFormatter()
    record = LogRecord("_test_", 20, "module.py", 20, "Message", (), None)
    json_string = formatter.format(record)
    record_dict = json.loads(json_string)
    assert (len(record_dict.keys())) == 5



"""Knowledge base for CPU, OS, and Platform informations.

Note that even if this is pure data this is not stored as a yaml. This
knowledge base is very often read and the cost of parsing the data can
be significant in some context.
"""
from __future__ import annotations

CPU_INFO = {
    "aarch64": {"endian": "little", "bits": 64},
    "arm": {"endian": "little", "bits": 32},
    "avr": {"endian": "little", "bits": 16},
    "powerpc": {"endian": "big", "bits": 32},
    "powerpc64": {"endian": "big", "bits": 64},
    "sparc": {"endian": "big", "bits": 32},
    "sparc64": {"endian": "big", "bits": 64},
    "x86": {"endian": "little", "bits": 32},
    "x86_64": {"endian": "little", "bits": 64},
}

OS_INFO = {
    "aix": {"is_bareboard": False, "version": "5.2", "exeext": "", "dllext": ".so"},
    "darwin": {
        "is_bareboard": False,
        "version": "9.6.0",
        "exeext": "",
        "dllext": ".dylib",
    },
    "freebsd": {
        "is_bareboard": False,
        "version": "unknown",
        "exeext": "",
        "dllext": ".so",
    },
    "openbsd": {
        "is_bareboard": False,
        "version": "unknown",
        "exeext": "",
        "dllext": ".so",
    },
    "netbsd": {
        "is_bareboard": False,
        "version": "unknown",
        "exeext": "",
        "dllext": ".so",
    },
    "dragonfly": {
        "is_bareboard": False,
        "version": "unknown",
        "exeext": "",
        "dllext": ".so",
    },
    "ios": {
        "is_bareboard": False,
        "version": "unknown",
        "exeext": "",
        "dllext": ".dylib",
    },
    "linux": {
        "is_bareboard": False,
        "version": "unknown",
        "exeext": "",
        "dllext": ".so",
    },
    "solaris": {"is_bareboard": False, "version": "2.8", "exeext": "", "dllext": ".so"},
    "windows": {
        "is_bareboard": False,
        "version": "XP",
        "exeext": ".exe",
        "dllext": ".dll",
    },
    "none": {"is_bareboard": True, "version": "unknown", "exeext": "", "dllext": ""},
}

PLATFORM_INFO = {
    "aarch64-elf": {"cpu": "aarch64", "os": "none", "is_hie": True},
    "aarch64-ios": {"cpu": "aarch64", "os": "ios", "is_hie": False},
    "aarch64-linux": {"cpu": "aarch64", "os": "linux", "is_hie": False},
    "arm-android": {"cpu": "arm", "os": "android", "is_hie": False},
    "arm-elf": {"cpu": "arm", "os": "none", "is_hie": True},
    "arm-ios": {"cpu": "arm", "os": "ios", "is_hie": False},
    "avr-elf": {"cpu": "avr", "os": "none", "is_hie": True},
    "arm-linux": {"cpu": "arm", "os": "linux", "is_hie": False, "endian": "little"},
    "ppc-aix": {"cpu": "powerpc", "os": "aix", "is_hie": False},
    "ppc-linux": {"cpu": "powerpc", "os": "linux", "is_hie": False},
    "raspberrypi-linux": {
        "cpu": "arm",
        "os": "linux",
        "is_hie": False,
        "endian": "little",
    },
    "sparc64-solaris": {"cpu": "sparc64", "os": "solaris", "is_hie": False},
    "sparc-solaris": {"cpu": "sparc", "os": "solaris", "is_hie": False},
    "x86_64-linux": {"cpu": "x86_64", "os": "linux", "is_hie": False},
    "x86_64-darwin": {"cpu": "x86_64", "os": "darwin", "is_hie": False},
    "x86-freebsd": {"cpu": "x86", "os": "freebsd", "is_hie": False},
    "x86-openbsd": {"cpu": "x86", "os": "openbsd", "is_hie": False},
    "x86-netbsd": {"cpu": "x86", "os": "netbsd", "is_hie": False},
    "x86-dragonfly": {"cpu": "x86", "os": "dragonfly", "is_hie": False},
    "x86_64-freebsd": {"cpu": "x86_64", "os": "freebsd", "is_hie": False},
    "x86_64-openbsd": {"cpu": "x86_64", "os": "openbsd", "is_hie": False},
    "x86_64-netbsd": {"cpu": "x86_64", "os": "netbsd", "is_hie": False},
    "x86_64-dragonfly": {"cpu": "x86_64", "os": "dragonfly", "is_hie": False},
    "x86-linux": {"cpu": "x86", "os": "linux", "is_hie": False},
    "x86-solaris": {"cpu": "x86", "os": "solaris", "is_hie": False},
    "x86_64-solaris": {"cpu": "x86_64", "os": "solaris", "is_hie": False},
    "x86-windows": {"cpu": "x86", "os": "windows", "is_hie": False},
    "x86_64-windows": {"cpu": "x86_64", "os": "windows", "is_hie": False},
    "x86_64-windows64": {"cpu": "x86_64", "os": "windows", "is_hie": False},
}

BUILD_TARGETS = {
    "aarch64-elf": {"name": "aarch64-elf"},
    "aarch64-ios": {"name": "aarch64-apple-darwin"},
    "aarch64-linux": {"name": "aarch64-linux-gnu"},
    "arm-android": {"name": "arm-linux-androideabi"},
    "arm-elf": {"name": "arm-eabi"},
    "arm-ios": {"name": "arm-apple-darwin10"},
    "avr-elf": {"name": "avr"},
    "arm-linux": {"name": "arm-linux-gnueabi"},
    "ppc-aix": {"name": "powerpc-ibm-aix%(os_version)s.0.0"},
    "ppc-linux": {"name": "powerpc-generic-linux-gnu"},
    "raspberrypi-linux": {"name": "arm-linux-gnueabihf"},
    "sparc64-solaris": {"name": "sparc64-sun-solaris%(os_version)s"},
    "sparc-solaris": {"name": "sparc-sun-solaris%(os_version)s"},
    "x86_64-linux": {"name": "x86_64-pc-linux-gnu"},
    "x86_64-darwin": {"name": "x86_64-apple-darwin%(os_version)s"},
    "x86-freebsd": {"name": "i386-pc-freebsd%(os_version)s"},
    "x86-openbsd": {"name": "i386-pc-openbsd%(os_version)s"},
    "x86-netbsd": {"name": "i386-pc-netbsd%(os_version)s"},
    "x86-dragonfly": {"name": "i386-pc-dragonfly%(os_version)s"},
    "x86_64-freebsd": {"name": "x86_64-pc-freebsd%(os_version)s"},
    "x86_64-openbsd": {"name": "x86_64-pc-openbsd%(os_version)s"},
    "x86_64-netbsd": {"name": "x86_64-pc-netbsd%(os_version)s"},
    "x86_64-dragonfly": {"name": "x86_64-pc-dragonfly%(os_version)s"},
    "x86-linux": {"name": "i686-pc-linux-gnu"},
    "x86-solaris": {"name": "i686-pc-solaris%(os_version)s"},
    "x86_64-solaris": {"name": "x86_64-sun-solaris%(os_version)s"},
    "x86-windows": {"name": "i686-pc-mingw32"},
    "x86_64-windows": {"name": "x86_64-pc-mingw32"},
    "x86_64-windows64": {"name": "x86_64-w64-mingw32"},
}

# The following table is used to guess a product name from the output of
# uname on the host. Users of this data are expected to match the specified
# regular expressions against that output to find a matching key. Order is
# not significant so if a traversal matches multiple entries the one matched
# is undefined. It is critical therefore that the supplied expressions match
# only the intended product and no other values of uname potentially output
# by a different host. IMPORTANT: Systems that can be only used as target in
# cross context should not be added to that table.

HOST_GUESS = {
    # platform : OS (uname[0]), machine (uname[1]), proc (uname[4 or 5])
    "ppc-aix": {"os": "AIX", "cpu": None},
    "x86_64-darwin": {"os": "Darwin", "cpu": "i386"},
    "x86-freebsd": {"os": "FreeBSD", "cpu": "i386"},
    "x86-openbsd": {"os": "OpenBSD", "cpu": None},
    "x86-netbsd": {"os": "NetBSD", "cpu": None},
    "x86-dragonfly": {"os": "DragonFly", "cpu": None},
    "x86_64-freebsd": {"os": "FreeBSD", "cpu": "amd64"},
    "x86_64-openbsd": {"os": "OpenBSD", "cpu": None},
    "x86_64-netbsd": {"os": "NetBSD", "cpu": None},
    "x86_64-dragonfly": {"os": "DragonFly", "cpu": None},
    "ppc-linux": {"os": "Linux", "cpu": "powerpc.*|ppc64"},
    "x86-linux": {"os": "Linux", "cpu": "i.86|pentium"},
    "x86_64-linux": {"os": "Linux", "cpu": "x86_64"},
    "sparc-solaris": {"os": "SunOS", "cpu": "sparc"},
    "x86-solaris": {"os": "SunOS", "cpu": "i386"},
    "x86-windows": {"os": "Windows", "cpu": None},
}



from __future__ import annotations

from e3.error import E3Error

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional
    from e3.os.process import Run


class VCSError(E3Error):
    def __init__(self, message: str, origin: str, process: Optional[Run] = None):
        super().__init__(message, origin)
        self.origin = origin
        self.message = message
        self.process = process



"""Provides high level queries on Anod specs."""
from __future__ import annotations

import collections
from typing import overload, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple
    from e3.anod.spec import Anod
    from e3.anod.context import AnodContext
    from e3.anod.package import SourceBuilder

    class SourceKey(NamedTuple):
        anod_uid: str
        src_name: str
        publish: bool

    class PackageKey(NamedTuple):
        anod_uid: str
        track: bool
        has_closure: bool


else:
    SourceKey = collections.namedtuple("SourceKey", ["anod_uid", "src_name", "publish"])
    PackageKey = collections.namedtuple(
        "PackageKey", ["anod_uid", "track", "has_closure"]
    )


@overload
def get_build_node(
    anod_instance: Anod, context: AnodContext, default: None = None
) -> Optional[Anod]:
    ...


@overload
def get_build_node(anod_instance: Anod, context: AnodContext, default: Anod) -> Anod:
    ...


def get_build_node(
    anod_instance: Anod, context: AnodContext, default: Optional[Anod] = None
) -> Optional[Anod]:
    """Return the build anod instance corresponding to an install.

    :param anod_instance: an Anod instance
    :param context: Anod context
    :param default: value returned if the build Anod instance does not exist
    :return: the associated build anod instance. If such instance does not
        exist then return default
    """
    if anod_instance.kind != "install":
        return default

    elif anod_instance.uid[:-8] + ".build" in context.tree:
        return context[anod_instance.uid[:-8] + ".build"].anod_instance  # type: ignore
    else:
        return default


def get_source_builder(
    anod_instance: Anod, source_name: str, local_sources_only: bool = False
) -> Optional[SourceBuilder]:
    """Given a source name return the associated builder.

    :param anod_instance: an Anod instance
    :param source_name: a source name
    :param local_sources_only: if True consider only builders declared in the
        spec itself. if False also consider source builders in source
        dependencies (i.e: require='source_pkg')
    :return: a source builder or None if no builder can be found.
    """
    # First look locally
    builder = None

    if anod_instance.source_pkg_build is not None:
        builder = next(
            (b for b in anod_instance.source_pkg_build if b.name == source_name), None
        )

    if builder is None and not local_sources_only:
        # If needed look into the deps
        for dep in list(anod_instance.deps.values()):
            if dep.kind != "source":
                continue
            builder = get_source_builder(dep, source_name, local_sources_only=True)
            if builder is not None:
                break
    return builder


class SourceClosure:
    """Helper object to resolve source closure for a given spec instance."""

    def __init__(
        self,
        anod_instance: Anod,
        context: AnodContext,
        expand_packages: bool = False,
        data_key: Optional[Callable[[Any], str]] = None,
    ):
        """Initialize a SourceClosure.

        :param anod_instance: an Anod instance
        :param context: the Anod context in which anod_instance was created
        :param expand_packages: whether to attempt to find source closure for
            installation dependencies which are producing a package
        :param data_key: function that given a data associated to a source
            return a key. By default __hash__ is used.
        """
        self.anod_instance = anod_instance
        self.expand_packages = expand_packages
        self.context = context
        self.source_list: Dict[SourceKey, Optional[Any]] = {}
        self.package_list: Dict[PackageKey, Optional[Any]] = {}
        self.compute_closure(self.anod_instance, publish=True)
        self.data_key = data_key
        if self.data_key is None:
            self.data_key = lambda x: x.__hash__()

    def compute_closure(self, spec: Anod, publish: bool) -> None:
        """Compute the source closure (internal function).

        :param spec: an anod instance to inspect
        :param publish: whether the sources found should be marked internal
            or not.
        """
        # If this is an installation add the corresponding package
        if spec.kind == "install":
            has_closure = get_build_node(spec, self.context) is not None
            self.package_list[PackageKey(spec.uid, publish, has_closure)] = None
            return

        # Otherwise consider the sources and recursively iterate on the
        # dependencies.
        for source in getattr(spec, "%s_source_list" % spec.kind, []):
            publish_source = publish and source.publish
            self.source_list[SourceKey(spec.uid, source.name, publish_source)] = None

        # Follow dependencies
        for dep, dep_spec in list(self.context.dependencies[spec.uid].values()):

            # Only consider build and install dependency (discard source deps)
            if dep.kind in ("build", "install"):
                if dep_spec.kind == "install":
                    # An install dep should be expanded whenever
                    # expand_packages is defined or when there is no component
                    # information generated (i.e: when dep_spec.component is
                    # None)
                    if self.expand_packages or dep_spec.component is None:
                        dep_spec = get_build_node(
                            dep_spec, context=self.context, default=dep_spec
                        )
                self.compute_closure(dep_spec, publish and dep.track)

    def resolve_package(self, spec_uid: str, data: List[Tuple[Any, bool]]) -> None:
        """Associate source information to a given package.

        :param spec_uid: the anod uid
        :param data: list of data associated to the package. This is a list of
            tuples (src_data, publish). Where src_data is the data for a given
            source package and publish a boolean indicated whether the entry is
            internal or not.
        """
        for pkg_key in self.package_list:
            if pkg_key.anod_uid == spec_uid:
                self.package_list[pkg_key] = data

    def resolve_source(self, source_name: str, data: Any) -> None:
        """Associate source information to a given source.

        :param source_name: the source name. The closure resolution is
            done locally so we can assume that for all occurences of a
            given source name the same data should be associated
        :param data: data associated with the source
        """
        for src_key in self.source_list:
            if src_key.src_name == source_name:
                self.source_list[src_key] = data

    def get_source_list(self) -> List[List[Any]]:
        """Get the closure source list.

        The function return the list of data for the sources in the closure.
            Note that if a package present in the closure is both marked as
            untracked and that closure cannot be found then it is ignored.
            For any other source or package if no data is associated then
            an assert exception will be raised

        :return: a list of list (source, publish)
        """
        result: Dict[str, List[Any]] = {}
        for src_key, src in list(self.source_list.items()):
            assert src is not None, "missing resolution"
            if TYPE_CHECKING:
                assert self.data_key is not None
            result_key = self.data_key(src)
            result[result_key] = [
                src,
                result.get(result_key, (None, False))[1] or src_key.publish,
            ]

        for pkg_key, pkg in list(self.package_list.items()):
            if not pkg_key.has_closure and not pkg_key.track:
                continue
            assert pkg is not None, "missing resolution"

            for src, publish in pkg:
                if TYPE_CHECKING:
                    assert self.data_key is not None
                result_key = self.data_key(src)

                result[result_key] = [
                    src,
                    result.get(result_key, (None, False))[1]
                    or (publish and pkg_key.track),
                ]
        return list(result.values())



"""Fingerprints handling.

Fingerprints objects provides a synthetic view of a set of elements.
The purpose is to allow users to build fingerprints based on that
set of elements, and then use fingerprint comparison as fast method
for determining whether the same set of elements might have changed
since the last time those elements were checked.

One possible usage for fingerprints is determine whether something
we built is still up to date, or should be rebuilt. For that,
what we would do at the end of a successful build is we create
a fingerprint using elements such as the sources we used to perform
the build, the dependencies for our build, the version of the compiler,
etc etc. Later on, when asking ourselves whether we need to re-build
or not, we would compute a new fingerprint using the same elements,
but updated to the current situation, and compare it with the one
we previously computed. If different, we should rebuild.
"""


from __future__ import annotations

import hashlib
import json
import os

from typing import TYPE_CHECKING

import e3.log
from e3.env import Env
from e3.error import E3Error
from e3.fs import get_filetree_state
from e3.hash import sha256

if TYPE_CHECKING:
    from typing import Dict, Optional, Set

logger = e3.log.getLogger("fingerprint")

FINGERPRINT_VERSION = "1.2"
# This value should be bumped each time computation of the fingerprint changes.
# This ensures we don't try to compare fingerprints with different meanings.


class Fingerprint:
    """Fingerprint class.

    :ivar elements: a dictionary containing the checksum/id for each element
        part of the fingerprint. The key a string identifying the
        element.
    """

    def __init__(self) -> None:
        """Initialise a new fingerprint instance."""
        self.elements: Dict[str, str] = {}
        self.add("os_version", Env().build.os.version)
        # ??? add more detailed information about the build machine so that
        # even a minor OS upgrade trigger a rebuild

        self.add("fingerprint_version", FINGERPRINT_VERSION)

    def add(self, name: str, value: str) -> None:
        """Add a fingerprint element.

        :param name: name of the new element
        :param value: associated value (should be a string)
        :raise: E3Error
        """
        if isinstance(value, str):
            self.elements[name] = value
        else:
            raise E3Error(
                f"value for {name} should be a string got {value}", "fingerprint.add",
            )

    def add_dir(self, path: str) -> None:
        """Add a file tree to the fingerprint.

        :param path: a path to a directory
        """
        assert os.path.isdir(path), "directory %s does not exist" % path
        self.elements[os.path.basename(path)] = get_filetree_state(path)

    def add_file(self, filename: str) -> None:
        """Add a file element to the fingerprint.

        :param filename: a path

        Adding a filename element to a fingerprint is equivalent to do add
        an element for which key is the basename of the file and value is
        is the sha256 of the content
        """
        assert os.path.isfile(filename), "filename %s does not exist" % filename
        self.elements[os.path.basename(filename)] = sha256(filename)

    def __eq__(self, other: object) -> bool:
        """Implement == operator for two fingerprints.

        :param other: object to compare with

        Two fingerprints are considered equals if both arguments are
        fingerprints and all elements of the fingerprint are equal
        """
        if not isinstance(other, Fingerprint):
            # The argument is not a fingerprint so return False
            return False

        if set(self.elements.keys()) != set(other.elements.keys()):
            # If the set of keys for each fingerprint is not equal
            # return False
            return False

        for k in self.elements:
            if self.elements[k] != other.elements[k]:
                return False

        return True

    def __ne__(self, other: object) -> bool:
        """Implement != operator for two fingerprints.

        See __eq__ functions.
        """
        return not self == other

    def compare_to(self, other: object) -> Optional[Dict[str, Set[str]]]:
        """Compare two fingerprints and return the differences.

        :return: a dictionary that list the differences or None if the two
          Fingerprint are equals. The returned dictionary contains three
          items. 'updated' list the elements that are in both fingerprints but
          that are different, 'obsolete' list the elements that are only in
          self, and 'new' the elements that are only in other
        :raise AssertError: if other is not a Fingerprint
        """
        # If other is None behave as if the previous fingerprint was an empty
        # fingerprint
        if other is None:
            other = Fingerprint()

        assert isinstance(
            other, Fingerprint
        ), "can compare only with Fingerprint objects"

        self_key_set = set(self.elements.keys())
        other_key_set = set(other.elements.keys())

        obsolete = self_key_set - other_key_set
        new = other_key_set - self_key_set

        # Compare common elements
        updated = set()
        for k in self_key_set & other_key_set:
            if self.elements[k] != other.elements[k]:
                updated.add(k)

        if len(updated) == 0 and len(new) == 0 and len(obsolete) == 0:
            return None
        else:
            return {"updated": updated, "new": new, "obsolete": obsolete}

    def __str__(self) -> str:
        """Return a string representation of the fingerprint."""
        return "\n".join(
            ["{}: {}".format(k, self.elements[k]) for k in sorted(self.elements.keys())]
        )

    def checksum(self) -> str:
        """Return the fingerprint's checksum.

        At the moment, the fingerprint uses the SHA256 hashing algorithm
        to compute the checksum.

        The function ensures that if two fingerprints are equal then
        the returned checksum for each of the fingerprint is equal.
        """
        key_list = list(self.elements.keys())
        key_list.sort()
        checksum = hashlib.sha256()
        for key in key_list:
            for chunk in (key, self.elements[key]):
                checksum.update(chunk.encode("utf-8"))
        return checksum.hexdigest()

    def save_to_file(self, filename: str) -> None:
        """Save the fingerprint to the given file.

        :param filename: The name of the file where to save the fingerprint.
        """
        # Save the fingerprint in a format that makes it easy for us
        # to support multiple fingerprint versions.
        #
        # For instance, while we have no metadata we want to save today
        # (eg: an attribute of the object), we can easily add a new
        # 'metadata' entry as a dictionary, with each element being
        # one piece of metadata we want to save.
        data = {"fingerprint_version": FINGERPRINT_VERSION, "elements": self.elements}

        with open(filename, "w") as f:
            json.dump(data, f, indent=2)

    @classmethod
    def load_from_file(cls, filename: str) -> Optional[Fingerprint]:
        """Return the fingerprint saved in the given file.

        Return None in the following situations:
            - The file does not contain a fingerprint we recognize;
            - The fingerprint has a version number we do not support.
            - If the fingerprint file does not exist

        :param filename: The name of the file where to load the fingerprint
            from.
        """
        if not os.path.isfile(filename):
            return None

        with open(filename) as f:
            try:
                data = json.load(f)
            except ValueError as e:
                logger.error(
                    "`%s' is not a properly formatted fingerprint file (%s)",
                    filename,
                    e,
                )
                return None

        bad_contents = None
        if not isinstance(data, dict):
            bad_contents = "not a dictionary"
        elif "fingerprint_version" not in data:
            bad_contents = '"fingerprint_version" key missing'
        elif "elements" not in data:
            bad_contents = '"elements" key missing'
        if bad_contents is not None:
            logger.error("`%s' is not a fingerprint file (%s)", filename, bad_contents)
            return None

        if data["fingerprint_version"] != FINGERPRINT_VERSION:
            logger.info(
                "Unsupported fingerprint version: %s", data["fingerprint_version"]
            )
            return None

        fingerprint = Fingerprint()
        fingerprint.elements = data["elements"]

        return fingerprint



from __future__ import annotations

import abc
import os
from typing import TYPE_CHECKING

import e3.log
from e3.anod.context import AnodContext
from e3.anod.loader import AnodSpecRepository
from e3.anod.sandbox import SandBox, SandBoxError
from e3.anod.sandbox.main import main
from e3.anod.spec import check_api_version
from e3.electrolyt.plan import Plan, PlanContext
from e3.electrolyt.run import ElectrolytJobFactory
from e3.env import BaseEnv
from e3.fs import mkdir
from e3.vcs.git import GitRepository

if TYPE_CHECKING:
    from typing import NoReturn

logger = e3.log.getLogger("e3.anod.SandBox")


class SandBoxAction(metaclass=abc.ABCMeta):

    require_sandbox = True

    def __init__(self, subparsers):
        self.parser = subparsers.add_parser(self.name, help=self.help)
        self.parser.set_defaults(action=self.name)
        self.add_parsers()
        if self.require_sandbox:
            self.parser.add_argument(
                "sandbox", help="path to the sandbox root directory"
            )

    @abc.abstractproperty
    def name(self) -> str:
        """Return the action name."""
        pass  # all: no cover

    @abc.abstractproperty
    def help(self) -> str:
        """Return the help string associated with this action."""
        pass  # all: no cover

    @abc.abstractmethod
    def add_parsers(self):
        """Add new command line argument parsers."""
        pass  # all: no cover

    @abc.abstractmethod
    def run(self, args):
        """Run the action.

        :param args: command line arguments gotten with argparse.
        """
        pass  # all: no cover


class SandBoxCreate(SandBoxAction):

    name = "create"
    help = "Create a new sandbox"

    def add_parsers(self):
        self.parser.add_argument(
            "--spec-git-url",
            help="URL to retrieve Anod specification files, hosted in a git"
            " repository",
        )
        self.parser.add_argument(
            "--spec-git-branch",
            default="master",
            help="Name of the branch to checkout to get the Anod "
            "specification files from the repository defined by "
            "``--spec-git-url``",
        )

    def run(self, args):
        sandbox = SandBox()
        sandbox.root_dir = args.sandbox

        sandbox.create_dirs()

        if args.spec_git_url:
            g = GitRepository(sandbox.specs_dir)
            if e3.log.default_output_stream is not None:
                g.log_stream = e3.log.default_output_stream
            g.init()
            g.update(args.spec_git_url, args.spec_git_branch, force=True)

        sandbox.dump_configuration()
        sandbox.write_scripts()


class SandBoxShowConfiguration(SandBoxAction):

    name = "show-config"
    help = "Display sandbox configuration"

    # List of configuration key to show
    keys = ("spec_git_url", "spec_git_branch", "sandbox")

    def add_parsers(self):
        pass

    def run(self, args):
        def warn_invalid(msg):
            print(
                "the configuration is invalid, the argument parser got the"
                " following error:"
            )
            print(msg)

        sandbox = SandBox()
        sandbox.root_dir = args.sandbox

        try:
            cmd_line = sandbox.get_configuration()["cmd_line"]
        except (TypeError, KeyError) as msg:
            warn_invalid(msg)
            return

        argument_parser = main(get_argument_parser=True)

        def error(message: str) -> NoReturn:
            raise SandBoxError(message)

        argument_parser.error = error
        try:
            args = argument_parser.parse_args(cmd_line[2:])
        except SandBoxError as msg:
            warn_invalid(msg)
            return

        for k, v in vars(args).items():
            if k in self.keys:
                print(f"{k} = {v}")


class SandBoxExec(SandBoxCreate):

    name = "exec"
    help = "Execute anod action in an sandbox"

    def add_parsers(self):
        super().add_parsers()
        self.parser.add_argument("--specs-dir", help="Alternate spec directory to use")
        self.parser.add_argument(
            "--create-sandbox", action="store_true", help="Create the sandbox if needed"
        )
        self.parser.add_argument("--plan", metavar="FILE", help="Path to the plan")
        self.parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Show action item without any execution",
        )
        self.parser.add_argument("--resolver", help="Use specific resolver")

    def run(self, args):
        sandbox = SandBox()
        sandbox.root_dir = args.sandbox

        if args.specs_dir:
            sandbox.specs_dir = args.specs_dir

        if args.create_sandbox:
            sandbox.create_dirs()

        if args.create_sandbox and args.spec_git_url:
            mkdir(sandbox.specs_dir)
            g = GitRepository(sandbox.specs_dir)
            if e3.log.default_output_stream is not None:
                g.log_stream = e3.log.default_output_stream
            g.init()
            g.update(args.spec_git_url, args.spec_git_branch, force=True)

        sandbox.dump_configuration()
        sandbox.write_scripts()

        asr = AnodSpecRepository(sandbox.specs_dir)
        check_api_version(asr.api_version)

        # Load plan content if needed
        if args.plan:
            if not os.path.isfile(args.plan):
                raise SandBoxError(
                    "plan file %s does not exist" % args.plan, origin="SandBoxExec.run"
                )
            with open(args.plan) as plan_fd:
                plan_content = ["def main_entry_point():"]
                plan_content += [
                    "    %s" % line for line in plan_fd.read().splitlines()
                ]
                plan_content = "\n".join(plan_content)

            env = BaseEnv()
            cm = PlanContext(server=env)
            store = None
            resolver = getattr(
                AnodContext,
                str(args.resolver),
                AnodContext.always_create_source_resolver,
            )
            logger.debug("Using resolver %s", resolver.__name__)

            # Declare available actions and their signature
            def anod_action(module, build=None, host=None, target=None, qualifier=None):
                pass  # all: no cover

            for a in ("anod_install", "anod_build", "anod_test"):
                cm.register_action(a, anod_action)

            # Load the plan and execute
            plan = Plan(data={})
            plan.load_chunk(plan_content)
            actions = cm.execute(plan, "main_entry_point")

            ac = AnodContext(asr, default_env=env)
            for action in actions:
                ac.add_anod_action(
                    action.module,
                    action,
                    action.action.replace("anod_", "", 1),
                    action.qualifier,
                )

            # Check if machine plan is locally schedulable
            action_list = ac.schedule(resolver)
            e = ElectrolytJobFactory(sandbox, asr, store, dry_run=args.dry_run)
            e.run(action_list)



from __future__ import annotations

import ast
import inspect
import types
from datetime import datetime, timezone
from functools import partial
from typing import TYPE_CHECKING

from e3.collection.toggleable_bool import ToggleableBooleanGroup
from e3.electrolyt.entry_point import Machine, entry_point
from e3.env import BaseEnv
from e3.error import E3Error

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, Optional
    from e3.collection.toggleable_bool import ToggleableBoolean
    from e3.electrolyt.entry_point import EntryPoint


class PlanError(E3Error):
    """Error when parsing or executing the plan."""

    pass


class Plan:
    """Electrolyt Plan.

    :ivar entry_points: list of entry points found in the plans
    :vartype entry_points: dict
    """

    def __init__(
        self,
        data: Dict[str, Any],
        entry_point_cls: Optional[Dict[str, Callable[..., EntryPoint]]] = None,
        plan_ext: str = ".plan",
    ):
        """Initialize a new plan.

        :param data: a dictionary defining additional globals to push
            into plan associated module
        :param entry_point_cls: dict associating a list of decorator name
            with an entry point class
        :param plan_ext: plan extension, by default ".plan". This is used to
            detect whether a specific frame is in a plan or in our code. See
            PlanContext._add_action
        """
        self.mod = types.ModuleType("_anod_plan_")

        # Some additional user symbols
        for k, v in data.items():
            self.mod.__dict__[k] = v

        self.entry_points: Dict[str, EntryPoint] = {}

        if entry_point_cls is None:
            entry_point_cls = {}

        self.plan_ext = plan_ext

        self.mod.__dict__["machine"] = partial(
            entry_point, self.entry_points, Machine, "machine"
        )

        for name, cls in entry_point_cls.items():
            self.mod.__dict__[name] = partial(entry_point, self.entry_points, cls, name)

        self.plan_date = datetime.now(timezone.utc)
        self.mod.__dict__["cond"] = self.cond
        self.toggleable_bool_group = ToggleableBooleanGroup()

    def cond(self, name: str, date: Callable[[datetime], bool]) -> ToggleableBoolean:
        """Generate a new conditional boolean.

        :param name: variable name
        :param date: function that takes the plan date and return a boolean.
            This can be used to set a value depending on the day of the week,
            e.g. by setting the constant to True on weekend:
            lambda d: d.isoweekday() in [6, 7]
        """
        return self.toggleable_bool_group.add(name, date(self.plan_date))

    def load(self, filename: str) -> None:
        """Load python code from file.

        :param filename: path to the python code
        """
        with open(filename, "rb") as fd:
            source_code = fd.read()
        self.load_chunk(source_code, filename)

    def check(self, code_ast: ast.AST) -> None:
        """Check plan coding style."""
        del self, code_ast

    def load_chunk(self, source_code: bytes, filename: str = "<unknown>") -> None:
        """Load a chunk of Python code.

        :param source_code: python source code
        :param filename: filename associated with the Python code
        """
        code_ast = ast.parse(source_code, filename)
        self.check(code_ast)
        code = compile(code_ast, filename, "exec")
        exec(code, self.mod.__dict__)


class PlanContext:
    """Context in which a Plan is executed."""

    def __init__(
        self,
        stack: Optional[List[BaseEnv]] = None,
        plan: Optional[Plan] = None,
        ignore_disabled: bool = True,
        server: Optional[BaseEnv] = None,
        build: Optional[str] = None,
        host: Optional[str] = None,
        target: Optional[str] = None,
        enabled: bool = True,
        **kwargs: Any,
    ):
        """Initialize an execution context or a scope.

        :param stack: stack of BaseEnv object that keep track of scopes. Used
            only internally. User instantiation of PlanContext should be done
            with stack set to None.
        :param plan: the plan to execute
        :param ignore_disabled: when true, discard all lines in
            blocks "with defaults(enabled=False):"
        :param server: a BaseEnv object that represent the host default env.
            server parameter is taken into account only during creation of
            the initial context.
        :param build: see e3.env.BaseEnv.set_env
        :param host: see e3.env.BaseEnv.set_env
        :param target: see e3.env.BaseEnv.set_env
        :param enabled: whether the plan line is enabled or disabled
        :param kwargs: additional data for the current scope/context
        """
        self.ignore_disabled = ignore_disabled
        if stack:
            # This is a scope creation, so copy current env
            self.stack = stack
            self.plan = plan
            new = self.stack[-1].copy(build=build, host=host, target=target)

            if not enabled:
                # we are in a block with enabled=False set, disable all
                # lines in that block
                new.enabled = enabled

        else:
            self.stack = []
            if server is not None:
                # This is a new context
                new = server.copy(build, host, target)
            else:
                # This is a new context with no server information. In that
                # case retrieve defaults for the local machine.
                new = BaseEnv()
                new.set_env(build, host, target)
            new.enabled = enabled

        # Store additional data
        for k, v in kwargs.items():
            setattr(new, k, v)

        # And push on the stack
        self.stack.append(new)

        # Registered functions that correspond to actions. Note that
        # there is no need to propagate registered action to children
        # scopes. Only initial context use them. The overall scheme
        # works also because all scopes refer to the same stack of env
        # (because the object is mutable).
        self.actions: Dict[str, Callable] = {}
        self.action_list: List[BaseEnv] = []

    def register_action(self, name: str, fun: Callable) -> None:
        """Register a function that correspond to an action.

        :param name: name used in the plans
        :param fun: python function. The function itself does
            not require an implementation. Only signature is
            is used
        """
        self.actions[name] = fun

    @property
    def env(self) -> BaseEnv:
        """Get environment for current scope.

        :return: the current scope environment
        """
        return self.stack[-1]

    @property
    def default_env(self) -> BaseEnv:
        """Get initial environment.

        :return: the environment set during creation of the
            initial context
        """
        return self.stack[0]

    def execute(
        self, plan: Plan, entry_point_name: str, verify: bool = False
    ) -> List[BaseEnv]:
        """Execute a plan.

        :param plan: the plan to execute
        :param entry_point_name: entry point to call in the plan. It can be
            either a function name in the plan or an entry_point function
        :param verify: verify whether the entry point name is a
            electrolyt entry point
        :raise: PlanError
        :return: a list of plan actions
        """
        # Give access to some useful data during plan execution
        plan.mod.__dict__["env"] = self.env
        plan.mod.__dict__["default_env"] = self.default_env

        defaults = partial(PlanContext, self.stack, plan, self.ignore_disabled)
        plan.mod.__dict__["defaults"] = defaults

        for a in self.actions:
            # On each action _add_action will be called with first parameter
            # fixed to the corresponding action name.
            fun = partial(self._add_action, a)
            plan.mod.__dict__[a] = fun

        self.action_list = []
        self.plan = plan

        if entry_point_name in plan.entry_points:
            plan.entry_points[entry_point_name].execute()
        else:
            # ??? An error should be raised as soon as entry points are
            # used everywhere
            plan.mod.__dict__[entry_point_name]()

        return self.action_list

    def _add_action(self, name: str, *args: Any, **kwargs: Any) -> None:
        """Process action calls in plans.

        :param name: action name
        :param args: positional arguments of the action call
        :param kwargs: keyword arguments of the action call
        """
        if self.ignore_disabled and not self.env.enabled:
            return

        if TYPE_CHECKING:
            assert self.plan is not None

        # ??? sometimes to understand the context it would be better to have
        # several lines of plan, e.g. when an action is created by calling
        # a function defined in a plan. Include all these lines separated
        # by ';'. A better fix would be to change plan_line to plan_lines
        # containing a tuple of plan_line.
        plan_line = "unknown filename:unknown lineno"
        # Retrieve the plan line
        try:
            caller_frames = inspect.getouterframes(frame=inspect.currentframe())
            caller_frames_in_plan = []
            for frame in caller_frames:
                frame_filename = frame.filename
                if frame_filename.endswith(self.plan.plan_ext):
                    caller_frames_in_plan.append(frame)
        except Exception:  # defensive code
            # do not crash when inspect frames fails
            pass
        else:
            if not caller_frames_in_plan:
                # No information ?
                pass
            else:
                plan_line = ";".join(
                    f"{caller_frame.filename}:{caller_frame.lineno}"
                    for caller_frame in caller_frames_in_plan
                )

        # First create our initial object based on current scope env
        result = self.env.copy()

        # Process arguments
        call_args = inspect.getcallargs(self.actions[name], *args, **kwargs)

        # Push function arguments into the result object. If an arg value
        # is None then environment is not updated (handling of defaults
        # coming from environment). For the build, host and target arguments
        # a special processing is done to make the corresponding set_env
        # call on the result BaseEnv object
        platform: Dict[str, Optional[str]] = {
            "build": None,
            "host": None,
            "target": None,
        }

        # Likewise board argument is used to change only the machine name
        # of the target. ??? change name ???
        board = None

        for k, v in call_args.items():
            if k in platform:
                platform[k] = v
            elif k == "board":
                board = v
            elif v is not None or not hasattr(result, k):
                setattr(result, k, v)

        # Handle propagation of environment from the context
        if platform["host"] is None and result.is_canadian:
            platform["host"] = "host"
        if platform["target"] is None and result.is_cross:
            platform["target"] = "target"
        if platform["target"] == result.host.platform:
            # ??? This special case is temporary ???
            # the goal is avoid cross from a -> a which are
            # not current supported.
            # Plans should be updated to use target='host'
            # instead of target=env.host.platform
            platform["target"] = "host"

        result.set_env(**platform)

        # If necessary adjust target machine name
        if board is not None:
            result.set_target(
                result.target.platform,
                result.target.os.version,
                board,
                result.target.os.mode,
            )

        # Set action attribute (with action name)
        result.action = name
        result.plan_line = plan_line
        result.plan_args = result.to_dict()
        result.plan_call_args = call_args

        # Push the action in the current list
        self.action_list.append(result)

    def __enter__(self):
        self.plan.mod.__dict__["env"] = self.env
        return None

    def __exit__(self, _type, _value, _tb):
        del _type, _value, _tb
        self.stack.pop()
        self.plan.mod.__dict__["env"] = self.env



import os

from e3.store.cache.backends.filecache import FileCache


def test_cache():
    fc = FileCache({"cache_dir": os.path.join(os.getcwd(), "cache")})
    fc.set("a", 1)
    fc.get("a") == 1
    fc.set("b", 2, timeout=-2)
    assert fc.get("b", 3) == 3
    fc.clear()
    assert fc.get("a", 0) == 0

    fc.set("a", 1)
    assert fc.get("a") == 1
    assert fc.has_resource("a")
    assert "a" in fc
    fc.delete("a")
    assert fc.get("a", 0) == 0

    assert fc.set("d", "nok", {"will": ["not", "work"]}) is False



import os
import sys

from e3.anod.status import ReturnValue
from e3.collection.dag import DAG
from e3.fingerprint import Fingerprint
from e3.fs import rm
from e3.job import EmptyJob, Job, ProcessJob
from e3.job.walk import Walk

import pytest

# A directory where we have the equivalent of a sandbox; basically,
# a place where we store some information as we perform the actions
# of a given DAG.
SBX_DIR = os.path.join(os.getcwd(), "sbx")

# A place where to store fingerprints...
FINGERPRINT_DIR = os.path.join(SBX_DIR, "fingerprints")

# A directory where download nodes will get their files from.
STORE_DIR = os.path.join(SBX_DIR, "store")

# A place where to store anything else that we might need between
# two runs via the Walk class.
SBX_TMP_DIR = os.path.join(SBX_DIR, "tmp")

# By convention in this testcase, jobs whose UID start with this prefix
# will copy a file from STORE_DIR whose name is '<UID minus prefix>.txt'
# where <UID minus prefix> is the job's UID after stripping the dealing
# DOWNLOAD_JOB_UID_PREFIX, and place it in SBX_DIR.
#
# This is to simulate jobs that download sources from the store.
#
# For instance, if the job name is 'download.hello-src', then the name
# of the file in both STORE_DIR and SBX_TMP_DIR is hello-src.txt.
DOWNLOAD_JOB_UID_PREFIX = "download."


@pytest.fixture()
def setup_sbx(request):
    """Automatically create a (temporary) sandbox.

    That sandbox is created before each test gets executed, and
    a finalizer is put in place to automatically delete that
    directory upon test tear-down.
    """

    def delete_sbx():
        if os.path.exists(SBX_DIR):
            rm(SBX_DIR, True)

    request.addfinalizer(delete_sbx)

    delete_sbx()
    os.mkdir(SBX_DIR)
    os.mkdir(FINGERPRINT_DIR)
    os.mkdir(STORE_DIR)
    os.mkdir(SBX_TMP_DIR)


def job_source_basename(uid: str) -> str:
    """Return the name of a source corresponding to the give job.

    In our testcase environment, we will consider that, if a DAG's action
    (corresponding to the given Job ID) depends on some sources, those
    sources are in a file called "<uid>.txt" where uid is the action's
    Job ID. Not all actions need sources, so the filename returned
    may or may not exist. Or said differently: if the file exists,
    then the action depends on the "source" filename returned by this
    function; if it does not, then the action does not depend on sources.

    Same thing if a job is a "download" job, but with a slightly different
    naming scheme (see DOWNLOAD_JOB_UID_PREFIX above).

    :param uid: A unique Job ID.
    """
    if uid.startswith(DOWNLOAD_JOB_UID_PREFIX):
        uid = uid[len(DOWNLOAD_JOB_UID_PREFIX) :]
    return uid + ".txt"


def source_fullpath(uid: str) -> str:
    """Return the fullpath of a job's sources, if present.

    :param uid: A unique Job ID.
    """
    return os.path.join(SBX_TMP_DIR, job_source_basename(uid))


def source_store_fullpath(uid: str) -> str:
    """Return the given job's fullpath to its sources in the store.

    :param uid: A unique Job ID.
    """
    assert uid.startswith(DOWNLOAD_JOB_UID_PREFIX)
    return os.path.join(STORE_DIR, job_source_basename(uid))


class DryRunJob(EmptyJob):
    """A job that does nothing, created when in dry-run mode.

    This class is identical to the EmptyJob class, but having it allows
    us to differentiate between the EmptyJob objects that e3 creates,
    and these jobs we create when in dry-run mode.
    """


class DoNothingJob(Job):
    """A Job which inherits Job's implementation of the status attribute."""

    def run(self):
        pass


class ControlledJob(ProcessJob):
    """A ProcessJob for testing purposes.

    The process's behavior is configured by the suffix's UID (see
    the "cmd" line method for more details on this.
    """

    def __init__(self, uid, data, notify_end):
        super().__init__(uid, data, notify_end)
        self.run_count = 0

    def run(self):
        self.run_count += 1
        return super().run()

    @property
    def cmdline(self):
        result = [sys.executable, "-c"]
        if self.uid.endswith("bad"):
            result.append("import sys; sys.exit(1)")
        elif self.uid.endswith("notready:once") and self.run_count < 2:
            result.append("import sys; sys.exit(%d)" % ReturnValue.notready.value)
        elif self.uid.endswith("notready:always"):
            result.append("import sys; sys.exit(%d)" % ReturnValue.notready.value)
        elif self.uid.startswith(DOWNLOAD_JOB_UID_PREFIX):
            result.append(
                "import shutil; shutil.copyfile(r'%s', r'%s')"
                % (source_store_fullpath(self.uid), source_fullpath(self.uid))
            )
        else:
            result.append('print("Hello World")')
        return result


class SimpleWalk(Walk):
    dry_run_mode = False

    def __init__(self, actions):
        # The list of jobs (by UID) that have been requeued.
        self.saved_jobs = {}
        self.requeued = {}
        super().__init__(actions)

    def request_requeue(self, job):
        """Requeue the job is not already queued once."""
        # First record the number of times we've been asked to requeue
        # that job, and allow requeuing only twice.
        if job.uid not in self.requeued:
            self.requeued[job.uid] = 0
        self.requeued[job.uid] += 1
        return self.requeued[job.uid] < 3

    def create_job(self, uid, data, predecessors, notify_end):
        if self.dry_run_mode:
            job = DryRunJob(uid, data, notify_end, status=ReturnValue.force_skip)
        elif uid.endswith("do-nothing"):
            job = DoNothingJob(uid, data, notify_end)
        else:
            job = ControlledJob(uid, data, notify_end)
        return job

    def get_job(self, uid, data, predecessors, notify_end):
        # Normally, deriving classes of class Walk are not expected
        # to override this method. However, we need to do it here
        # as a way to record some information each time this method
        # is called, so as to be able to verify some aspects of
        # the class' behavior.
        job = super().get_job(uid, data, predecessors, notify_end)
        self.saved_jobs[job.uid] = job
        return job


class FingerprintWalk(SimpleWalk):
    @classmethod
    def fingerprint_filename(cls, uid):
        return os.path.join(FINGERPRINT_DIR, uid)

    def compute_fingerprint(self, uid, data, is_prediction=False):
        if "fingerprint_after_job" in uid and is_prediction:
            return None
        if "no_fingerprint" in uid:
            return None
        f = Fingerprint()
        for pred_uid in self.actions.get_predecessors(uid):
            pred_fingerprint = self.new_fingerprints[pred_uid]
            if pred_fingerprint is not None:
                f.add("pred:%s" % pred_uid, pred_fingerprint.checksum())
        if os.path.exists(source_fullpath(uid)):
            f.add_file(source_fullpath(uid))
        return f

    def save_fingerprint(self, uid, fingerprint):
        if self.dry_run_mode:
            # In dry-run mode, we don't do anything, so we should not
            # touch the fingerprint either.
            return

        filename = self.fingerprint_filename(uid)
        if fingerprint is None:
            if os.path.exists(filename):
                os.remove(filename)
        else:
            fingerprint.save_to_file(filename)

    def load_previous_fingerprint(self, uid):
        # In dry-run mode, the fingerprints on file are let untouched,
        # so they might be out of date compared to this job's status
        # as part of this dry run. So, if we have already computed
        # the fingerprint before, use that.
        if self.dry_run_mode and uid in self.new_fingerprints:
            return self.new_fingerprints[uid]

        filename = self.fingerprint_filename(uid)
        if os.path.exists(filename):
            return Fingerprint.load_from_file(filename)
        else:
            return None


class FingerprintWalkDryRun(FingerprintWalk):
    dry_run_mode = True


@pytest.mark.parametrize("walk_class", [SimpleWalk, FingerprintWalk])
class TestWalk:
    def test_good_job_no_predecessors(self, walk_class, setup_sbx):
        """Simple case of a leaf job."""
        actions = DAG()
        actions.add_vertex("1")
        c = walk_class(actions)

        job = c.saved_jobs["1"]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

        assert c.job_status == {"1": ReturnValue.success}
        assert c.requeued == {}

        # In the situation where we are using fingerprints,
        # verify the behavior when re-doing a walk with
        # the same DAG.

        if walk_class == FingerprintWalk:
            r2 = walk_class(actions)

            job = r2.saved_jobs["1"]
            assert isinstance(job, EmptyJob)
            assert job.should_skip is True
            assert job.status == ReturnValue.skip

            assert r2.job_status == {"1": ReturnValue.skip}
            assert r2.requeued == {}

    def test_bad_job_no_predecessors(self, walk_class, setup_sbx):
        """Simple case of a leaf job failing."""
        actions = DAG()
        actions.add_vertex("1.bad")
        c = walk_class(actions)

        job = c.saved_jobs["1.bad"]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue(1)
        assert c.job_status == {"1.bad": ReturnValue(1)}
        assert c.requeued == {}

        # In the situation where we are using fingerprints,
        # verify the behavior when re-doing a walk with
        # the same DAG.

        if walk_class == FingerprintWalk:
            r2 = walk_class(actions)

            job = r2.saved_jobs["1.bad"]
            assert isinstance(job, ControlledJob)
            assert job.should_skip is False
            assert job.status == ReturnValue(1)
            assert r2.job_status == {"1.bad": ReturnValue(1)}
            assert r2.requeued == {}

    def test_failed_predecessor(self, walk_class, setup_sbx):
        """Simulate the scenarior when a predecessor failed."""
        actions = DAG()
        actions.add_vertex("1.bad")
        actions.add_vertex("2", predecessors=["1.bad"])
        c = walk_class(actions)

        job = c.saved_jobs["1.bad"]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue(1)

        job = c.saved_jobs["2"]
        assert isinstance(job, EmptyJob)
        assert job.should_skip is True
        assert job.status == ReturnValue.force_fail

        assert c.job_status == {"1.bad": ReturnValue(1), "2": ReturnValue.force_fail}
        assert c.requeued == {}

        # In the situation where we are using fingerprints,
        # verify the behavior when re-doing a walk with
        # the same DAG.

        if walk_class == FingerprintWalk:
            r2 = walk_class(actions)

            job = r2.saved_jobs["1.bad"]
            assert isinstance(job, ControlledJob)
            assert job.should_skip is False
            assert job.status == ReturnValue(1)

            job = r2.saved_jobs["2"]
            assert isinstance(job, EmptyJob)
            assert job.should_skip is True
            assert job.status == ReturnValue.force_fail

            assert r2.job_status == {
                "1.bad": ReturnValue(1),
                "2": ReturnValue.force_fail,
            }
            assert r2.requeued == {}

    def test_job_not_ready_then_ok(self, walk_class, setup_sbx):
        """Rerunning a job that first returned notready."""
        actions = DAG()
        actions.add_vertex("1.notready:once")
        c = walk_class(actions)

        job = c.saved_jobs["1.notready:once"]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

        assert c.job_status == {"1.notready:once": ReturnValue.success}
        assert c.requeued == {"1.notready:once": 1}

        # In the situation where we are using fingerprints,
        # verify the behavior when re-doing a walk with
        # the same DAG.

        if walk_class == FingerprintWalk:
            r2 = walk_class(actions)

            job = r2.saved_jobs["1.notready:once"]
            assert isinstance(job, EmptyJob)
            assert job.should_skip is True
            assert job.status == ReturnValue.skip

            assert r2.job_status == {"1.notready:once": ReturnValue.skip}
            assert r2.requeued == {}

    def test_job_never_ready(self, walk_class, setup_sbx):
        """Trying to run a job repeatedly returning notready."""
        actions = DAG()
        actions.add_vertex("1.notready:always")
        c = walk_class(actions)

        job = c.saved_jobs["1.notready:always"]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.notready

        assert c.job_status == {"1.notready:always": ReturnValue.notready}
        assert c.requeued == {"1.notready:always": 3}

        # In the situation where we are using fingerprints,
        # verify the behavior when re-doing a walk with
        # the same DAG.

        if walk_class == FingerprintWalk:
            r2 = walk_class(actions)

            job = r2.saved_jobs["1.notready:always"]
            assert isinstance(job, ControlledJob)
            assert job.should_skip is False
            assert job.status == ReturnValue.notready

            assert r2.job_status == {"1.notready:always": ReturnValue.notready}
            assert r2.requeued == {"1.notready:always": 3}

    def test_do_nothing_job(self, walk_class, setup_sbx):
        """Test DAG leading us to create a DoNothingJob object."""
        actions = DAG()
        actions.add_vertex("1.do-nothing")
        actions.add_vertex("2", predecessors=["1.do-nothing"])
        c = walk_class(actions)

        job = c.saved_jobs["1.do-nothing"]
        assert isinstance(job, DoNothingJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

        job = c.saved_jobs["2"]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

        assert c.job_status == {
            "1.do-nothing": ReturnValue.success,
            "2": ReturnValue.success,
        }
        assert c.requeued == {}

        # In the situation where we are using fingerprints,
        # verify the behavior when re-doing a walk with
        # the same DAG.

        if walk_class == FingerprintWalk:
            r2 = walk_class(actions)

            job = r2.saved_jobs["1.do-nothing"]
            assert isinstance(job, EmptyJob)
            assert job.should_skip is True
            assert job.status == ReturnValue.skip

            job = r2.saved_jobs["2"]
            assert isinstance(job, EmptyJob)
            assert job.should_skip is True
            assert job.status == ReturnValue.skip

            assert r2.job_status == {
                "1.do-nothing": ReturnValue.skip,
                "2": ReturnValue.skip,
            }
            assert r2.requeued == {}


def test_source_deps(setup_sbx):
    """Try runs with source dependencies changing between runs."""
    actions = DAG()
    actions.add_vertex("1")
    actions.add_vertex("2", predecessors=["1"])
    actions.add_vertex("3")
    actions.add_vertex("4", predecessors=["2", "3"])
    actions.add_vertex("5")

    # Create source dependencies for each actions
    for uid in ("1", "2", "3", "4", "5"):
        with open(source_fullpath(uid), "w") as f:
            f.write("contents of sources for action %s\n" % uid)

    # Now, execute our planned actions

    r1 = FingerprintWalk(actions)

    for uid in ("1", "2", "3", "4", "5"):
        job = r1.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r1.job_status == {
        "1": ReturnValue.success,
        "2": ReturnValue.success,
        "3": ReturnValue.success,
        "4": ReturnValue.success,
        "5": ReturnValue.success,
    }
    assert r1.requeued == {}

    # Executing it again should result in all actions being skipped.

    r2 = FingerprintWalk(actions)

    for uid in ("1", "2", "3", "4", "5"):
        job = r2.saved_jobs[uid]
        assert isinstance(job, EmptyJob)
        assert job.should_skip is True
        assert job.status == ReturnValue.skip

    assert r2.job_status == {
        "1": ReturnValue.skip,
        "2": ReturnValue.skip,
        "3": ReturnValue.skip,
        "4": ReturnValue.skip,
        "5": ReturnValue.skip,
    }
    assert r2.requeued == {}

    # Now change the sources of action '5', and run the actions
    # again. Only '5' should be executed.

    with open(source_fullpath("5"), "a") as f:
        f.write("Some more sources\n")

    r3 = FingerprintWalk(actions)

    for uid in ("1", "2", "3", "4"):
        job = r3.saved_jobs[uid]
        assert isinstance(job, EmptyJob)
        assert job.should_skip is True
        assert job.status == ReturnValue.skip

    for uid in ("5",):
        job = r3.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r3.job_status == {
        "1": ReturnValue.skip,
        "2": ReturnValue.skip,
        "3": ReturnValue.skip,
        "4": ReturnValue.skip,
        "5": ReturnValue.success,
    }
    assert r3.requeued == {}

    # Executing it again should result in all actions being skipped.

    r4 = FingerprintWalk(actions)

    for uid in ("1", "2", "3", "4", "5"):
        job = r4.saved_jobs[uid]
        assert isinstance(job, EmptyJob)
        assert job.should_skip is True
        assert job.status == ReturnValue.skip

    assert r4.job_status == {
        "1": ReturnValue.skip,
        "2": ReturnValue.skip,
        "3": ReturnValue.skip,
        "4": ReturnValue.skip,
        "5": ReturnValue.skip,
    }
    assert r4.requeued == {}

    # Change the sources of '1' and '2'. We expect '1' and '2' to
    # be re-run, as well as their dependences ('4', in this case).

    for uid in ("1", "2"):
        with open(source_fullpath(uid), "a") as f:
            f.write("Additional information\n")

    r5 = FingerprintWalk(actions)

    for uid in ("3", "5"):
        job = r5.saved_jobs[uid]
        assert isinstance(job, EmptyJob)
        assert job.should_skip is True
        assert job.status == ReturnValue.skip

    for uid in ("1", "2", "4"):
        job = r5.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r5.job_status == {
        "1": ReturnValue.success,
        "2": ReturnValue.success,
        "3": ReturnValue.skip,
        "4": ReturnValue.success,
        "5": ReturnValue.skip,
    }
    assert r5.requeued == {}

    # Change the sources of '1'. We expect '1' be re-run, and
    # as a consequence of that, '2' and '4' also should be rerun.

    with open(source_fullpath("1"), "a") as f:
        f.write("# Small comment\n")

    r5 = FingerprintWalk(actions)

    for uid in ("3", "5"):
        job = r5.saved_jobs[uid]
        assert isinstance(job, EmptyJob)
        assert job.should_skip is True
        assert job.status == ReturnValue.skip

    for uid in ("1", "2", "4"):
        job = r5.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r5.job_status == {
        "1": ReturnValue.success,
        "2": ReturnValue.success,
        "3": ReturnValue.skip,
        "4": ReturnValue.success,
        "5": ReturnValue.skip,
    }
    assert r5.requeued == {}


def test_predecessor_with_no_fingerprint(setup_sbx):
    actions = DAG()
    actions.add_vertex("1")
    actions.add_vertex("2.no_fingerprint", predecessors=["1"])
    actions.add_vertex("3", predecessors=["2.no_fingerprint"])
    actions.add_vertex("4", predecessors=["3"])

    # Execute our planned actions for the first time...

    r1 = FingerprintWalk(actions)

    for uid in ("1", "2.no_fingerprint", "3", "4"):
        job = r1.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r1.job_status == {
        "1": ReturnValue.success,
        "2.no_fingerprint": ReturnValue.success,
        "3": ReturnValue.success,
        "4": ReturnValue.success,
    }
    assert r1.requeued == {}

    # Re-execute the plan a second time.  Because '2.no_fingerprint'
    # has no fingerprint, both '2.no_fingerprint', but also the node
    # that depends directly on it should be re-executed. Node '4,
    # on the other hand, should only be re-executed if Node "3"'s
    # fingerprint changed. The way things are set up in this testsuite
    # is such that the fingerprint remained the same, so '4' is not
    # expected to be re-run.

    r2 = FingerprintWalk(actions)

    for uid in ("1", "4"):
        job = r2.saved_jobs[uid]
        assert isinstance(job, EmptyJob)
        assert job.should_skip is True
        assert job.status == ReturnValue.skip
    for uid in ("2.no_fingerprint", "3"):
        job = r2.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r2.job_status == {
        "1": ReturnValue.skip,
        "2.no_fingerprint": ReturnValue.success,
        "3": ReturnValue.success,
        "4": ReturnValue.skip,
    }
    assert r2.requeued == {}


def test_dry_run(setup_sbx):
    """Simulate the use actions with "dry run" behavior."""
    actions = DAG()
    actions.add_vertex("1")
    actions.add_vertex("2", predecessors=["1"])

    # First run in dry-run mode. Both actions are turned into
    # empty jobs with a force_skip status.

    r1 = FingerprintWalkDryRun(actions)

    job = r1.saved_jobs["1"]
    assert isinstance(job, DryRunJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.force_skip

    job = r1.saved_jobs["2"]
    assert isinstance(job, DryRunJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.force_skip

    assert r1.job_status == {"1": ReturnValue.force_skip, "2": ReturnValue.force_skip}
    assert r1.requeued == {}

    # Try it again in dry-mode; we should get the same result.

    r2 = FingerprintWalkDryRun(actions)

    job = r2.saved_jobs["1"]
    assert isinstance(job, DryRunJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.force_skip

    job = r2.saved_jobs["2"]
    assert isinstance(job, DryRunJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.force_skip

    assert r2.job_status == {"1": ReturnValue.force_skip, "2": ReturnValue.force_skip}
    assert r2.requeued == {}

    # Now, get action '1' done in normal (non-dry-run) mode.

    one_only = DAG()
    one_only.add_vertex("1")

    r3 = FingerprintWalk(one_only)

    job = r3.saved_jobs["1"]
    assert isinstance(job, ControlledJob)
    assert job.should_skip is False
    assert job.status == ReturnValue.success

    assert r3.job_status == {"1": ReturnValue.success}
    assert r3.requeued == {}

    # Try again the original plam in dry-run mode.
    #
    # This time, since '1' has been actually completed in a previous
    # run, action '1' should be skipped, and action '2' should be
    # a dry-run job...

    r4 = FingerprintWalkDryRun(actions)

    job = r4.saved_jobs["1"]
    assert isinstance(job, EmptyJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.skip

    job = r4.saved_jobs["2"]
    assert isinstance(job, DryRunJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.force_skip

    assert r4.job_status == {"1": ReturnValue.skip, "2": ReturnValue.force_skip}
    assert r4.requeued == {}

    # Run it again, this time in normal (non-dry-run) mode.
    #
    # This time, action '2' gets scheduled for real...

    r5 = FingerprintWalk(actions)

    job = r5.saved_jobs["1"]
    assert isinstance(job, EmptyJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.skip

    job = r5.saved_jobs["2"]
    assert isinstance(job, ControlledJob)
    assert job.should_skip is False
    assert job.status == ReturnValue.success

    assert r5.job_status == {"1": ReturnValue.skip, "2": ReturnValue.success}
    assert r5.requeued == {}

    # One more time, in dry-run mode again.
    #
    # There is nothing to be done, so the results should be
    # both actions are skipped.

    r6 = FingerprintWalkDryRun(actions)

    job = r6.saved_jobs["1"]
    assert isinstance(job, EmptyJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.skip

    job = r6.saved_jobs["2"]
    assert isinstance(job, EmptyJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.skip

    assert r6.job_status == {"1": ReturnValue.skip, "2": ReturnValue.skip}
    assert r6.requeued == {}


def test_computing_fingerprint_after_job_done(setup_sbx):
    """Test case where the fingerprint for one job has to be computed late."""
    download_uid = DOWNLOAD_JOB_UID_PREFIX + "fingerprint_after_job"
    actions = DAG()
    actions.add_vertex(download_uid)
    actions.add_vertex("2", predecessors=[download_uid])

    # Create the contents of the file that the download_uid job
    # will be "downloading" from the store.
    with open(source_store_fullpath(download_uid), "w") as f:
        f.write("Hello world")

    # Now, execute the plan for the first time.
    # All actions should get executed.

    r1 = FingerprintWalk(actions)
    for uid in (download_uid, "2"):
        job = r1.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r1.job_status == {
        download_uid: ReturnValue.success,
        "2": ReturnValue.success,
    }
    assert r1.requeued == {}

    # Now, rerun the plan.
    #
    # Because we set things up so that the fingerprint of the download_uid
    # job cannot be computed before the job is executed, we should see
    # that job being rerun (despite the fact that, in the end, the file
    # it downloads is the same). However, because the source it downloads
    # hasn't changed, job '2' should be skipped.

    r2 = FingerprintWalk(actions)

    job = r2.saved_jobs[download_uid]
    assert isinstance(job, ControlledJob)
    assert job.should_skip is False
    assert job.status == ReturnValue.success

    job = r2.saved_jobs["2"]
    assert isinstance(job, EmptyJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.skip

    assert r2.job_status == {download_uid: ReturnValue.success, "2": ReturnValue.skip}
    assert r2.requeued == {}

    # Simulate the case where we re-run the plan after the source
    # being downloaded has changed. This time, we expect job '2'
    # to be re-executed.

    with open(source_store_fullpath(download_uid), "w") as f:
        f.write("Hello brave new world")

    r3 = FingerprintWalk(actions)
    for uid in (download_uid, "2"):
        job = r3.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r3.job_status == {
        download_uid: ReturnValue.success,
        "2": ReturnValue.success,
    }
    assert r3.requeued == {}

    # One more time, just for kicks, where we re-execute the plan
    # where the file being downloaded hasn't changed.  Just to make
    # things slightly different, we'll remove the source file
    # already downloaded. It shouldn't prevent us from realizing
    # that the sources have not change, and so '2' can be skipped.

    rm(source_fullpath(download_uid))
    assert not os.path.exists(source_fullpath(download_uid))

    r4 = FingerprintWalk(actions)

    job = r4.saved_jobs[download_uid]
    assert isinstance(job, ControlledJob)
    assert job.should_skip is False
    assert job.status == ReturnValue.success

    job = r4.saved_jobs["2"]
    assert isinstance(job, EmptyJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.skip

    assert r4.job_status == {download_uid: ReturnValue.success, "2": ReturnValue.skip}
    assert r4.requeued == {}


def test_job_depending_on_job_with_no_predicted_fingerprint_failed(setup_sbx):
    """Test case where job depends on failed job with late fingerprint."""
    actions = DAG()
    actions.add_vertex("fingerprint_after_job.bad")
    actions.add_vertex("2", predecessors=["fingerprint_after_job.bad"])

    r1 = FingerprintWalk(actions)
    assert (
        r1.compute_fingerprint("fingerprint_after_job.bad", None, is_prediction=True)
        is None
    )

    # Check the status of the first job ('fingerprint_after_job.bad').
    # It should be a real job that returned a failure.
    job = r1.saved_jobs["fingerprint_after_job.bad"]
    assert isinstance(job, ControlledJob)
    assert job.should_skip is False
    assert job.status == ReturnValue(1)

    # Check the status of the second job ('2'); because that job depends
    # on a job that failed, it should show that the job was skipped.
    job = r1.saved_jobs["2"]
    assert isinstance(job, EmptyJob)
    assert job.should_skip is True
    assert job.status == ReturnValue.force_fail

    # Check that no job was requeued.
    assert r1.requeued == {}


def test_corrupted_fingerprint(setup_sbx):
    """Test the case where a fingerprint somehow got corrupted."""
    actions = DAG()
    actions.add_vertex("1")
    actions.add_vertex("2", predecessors=["1"])
    actions.add_vertex("3")
    actions.add_vertex("4", predecessors=["2", "3"])
    actions.add_vertex("5", predecessors=["4"])
    actions.add_vertex("6")

    # Now, execute the plan a first time; everything should be run
    # and finish succesfullly.

    r1 = FingerprintWalk(actions)

    for uid in ("1", "2", "3", "4", "5", "6"):
        job = r1.saved_jobs[uid]
        assert isinstance(job, ControlledJob)
        assert job.should_skip is False
        assert job.status == ReturnValue.success

    assert r1.job_status == {
        "1": ReturnValue.success,
        "2": ReturnValue.success,
        "3": ReturnValue.success,
        "4": ReturnValue.success,
        "5": ReturnValue.success,
        "6": ReturnValue.success,
    }
    assert r1.requeued == {}

    # Now, corrupt the fingerprint of node '3', and then rerun
    # the scheduler... We expect the following:
    #  - The scheduler does _not_ crash ;-)
    #  - The fingerprint of node '3' gets discarded, and as a result
    #    it should be re-run again.
    #  - Since nothing changed in node "3"'s predecessors, the end
    #    result for node '3' should be the same, which means
    #    its fingerprint should be the same as before the corruption.
    #    As a result of that, nodes '4' and '5', which directly
    #    or indirectly depend on node '3', do not need to be rerun.

    with open(r1.fingerprint_filename("3"), "w") as f:
        f.write("{")

    r2 = FingerprintWalk(actions)

    job = r2.saved_jobs["3"]
    assert isinstance(job, ControlledJob)
    assert job.should_skip is False
    assert job.status == ReturnValue.success

    for uid in ("1", "2", "4", "5", "6"):
        job = r2.saved_jobs[uid]
        assert isinstance(job, EmptyJob), "job %s should be EmptyJob" % uid
        assert job.should_skip is True
        assert job.status == ReturnValue.skip

    # Verify also that the fingerprint corruption is gone.
    f3 = Fingerprint.load_from_file(r2.fingerprint_filename("3"))
    assert isinstance(f3, Fingerprint)



import os
import sys

import e3.anod.helper
import e3.anod.sandbox
import e3.env
import e3.fs
import e3.os.process
import e3.platform
import yaml
from e3.vcs.git import GitRepository

import pytest

PROLOG = r"""# prolog file loaded before all specs


def main():
    import yaml
    import os

    with open(os.path.join(
            __spec_repository.spec_dir, 'conf.yaml')) as f:
        conf = yaml.safe_load(f)
        __spec_repository.api_version = conf['api_version']
        __spec_repository.repos = conf['repositories']


main()
del main
"""


def create_prolog(prolog_dir, content=PROLOG):
    """Create prolog.py file on the fly to prevent checkstyle error."""
    if os.path.isfile(os.path.join(prolog_dir, "prolog.py")):
        return
    with open(os.path.join(prolog_dir, "prolog.py"), "w") as f:
        f.write(content)


@pytest.fixture
def git_specs_dir(git, tmpdir):
    """Create a git repo to check out specs from it."""
    del git  # skip or fail when git is not installed

    # Create a e3-specs git repository for all tests in this module
    specs_dir = str(tmpdir.mkdir("e3-specs-git-repo"))

    try:
        specs_source_dir = os.path.join(os.path.dirname(__file__), "specs")
        for spec_file in os.listdir(specs_source_dir):
            if os.path.isfile(os.path.join(specs_source_dir, spec_file)):
                e3.fs.cp(
                    os.path.join(specs_source_dir, spec_file),
                    os.path.join(specs_dir, spec_file),
                )
        # Put prolog file
        create_prolog(specs_dir)
        if os.path.isdir(os.path.join(specs_dir, ".git")):
            return
        g = GitRepository(specs_dir)
        g.init()
        g.git_cmd(["config", "user.email", '"test@example.com"'])
        g.git_cmd(["config", "user.name", '"test"'])
        g.git_cmd(["add", "-A"])
        g.git_cmd(["commit", "-m", "'add all'"])
        yield "file://%s" % specs_dir.replace("\\", "/")
    finally:
        e3.fs.rm(specs_dir, True)


def test_deploy_sandbox():
    sandbox_dir = os.getcwd()
    e3.os.process.Run(["e3-sandbox", "-v", "-v", "create", sandbox_dir], output=None)
    assert os.path.isdir("log")

    assert (
        "sandbox = %s" % sandbox_dir
        in e3.os.process.Run(["e3-sandbox", "show-config", sandbox_dir]).out
    )

    e3.fs.mkdir("specs")

    with open(os.path.join("specs", "a.anod"), "w") as fd:
        fd.write("from e3.anod.spec import Anod\n")
        fd.write("class A(Anod):\n")
        fd.write("    pass\n")

    assert (
        "no primitive download"
        in e3.os.process.Run([os.path.join("bin", "anod"), "download", "a"]).out
    )

    with open(os.path.join("specs", "b.anod"), "w") as fd:
        fd.write("from e3.anod.spec import Anod\n")
        fd.write("class B(Anod):\n\n")
        fd.write("    @Anod.primitive()\n")
        fd.write("    def download(self):\n")
        fd.write("        pass\n")

    assert (
        "no download metadata returned by the download primitive"
        in e3.os.process.Run([os.path.join("bin", "anod"), "download", "b"]).out
    )


def test_sandbox_show_config_err():
    """Check e3-sandbox show-config with invalid sandbox."""
    sandbox_dir = os.getcwd()
    sandbox_conf = os.path.join("meta", "sandbox.yaml")
    e3.os.process.Run(["e3-sandbox", "-v", "-v", "create", sandbox_dir], output=None)
    with open(sandbox_conf) as f:
        conf = yaml.safe_load(f.read())
    conf["cmd_line"].append("--wrong-arg")
    with open(sandbox_conf, "w") as f:
        yaml.safe_dump(conf, stream=f)

    assert (
        "the configuration is invalid"
        in e3.os.process.Run(["e3-sandbox", "show-config", sandbox_dir]).out
    )

    with open(sandbox_conf, "w") as f:
        f.write("invalid")

    assert (
        "the configuration is invalid"
        in e3.os.process.Run(["e3-sandbox", "show-config", sandbox_dir]).out
    )


def test_sandbox_env():
    os.environ["GPR_PROJECT_PATH"] = "/foo"
    sandbox = e3.anod.sandbox.SandBox()
    sandbox.set_default_env()
    assert os.environ["GPR_PROJECT_PATH"] == ""


def test_sandbox_rootdir():
    sandbox = e3.anod.sandbox.SandBox()
    with pytest.raises(e3.anod.sandbox.SandBoxError):
        assert sandbox.root_dir

    sandbox.root_dir = "foo"
    sandbox.root_dir = "foo"
    assert os.path.relpath(sandbox.tmp_dir, sandbox.root_dir) == "tmp"
    assert (
        os.path.relpath(
            sandbox.get_build_space("bar").root_dir,
            os.path.join("foo", e3.env.Env().platform),
        )
        == "bar"
    )


def test_sandbox_create_git(git_specs_dir):
    """Check if sandbox create can load the specs from a git repo."""
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")
    with_git = e3.os.process.Run(
        ["e3-sandbox", "-v", "create", "--spec-git-url", git_specs_dir, sandbox_dir],
        output=None,
    )
    assert with_git.status == 0
    # Test structure
    for test_dir in ["bin", "log", "meta", "patch", "specs", "src", "tmp", "vcs"]:
        assert os.path.isdir(os.path.join(sandbox_dir, test_dir))
    # Test specs files if created
    specs_files = ["anod.anod", "dummyspec.anod", "conf.yaml", "prolog.py"]
    for filename in specs_files:
        assert os.path.isfile(os.path.join(sandbox_dir, "specs", filename))


def test_sandbox_exec_missing_spec_dir(git_specs_dir):
    """Test sandbox exec exception.

    - Check if sandbox exec raises exception if spec-dir is missing
    """
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)

    # Specs dir is missing
    no_specs = e3.os.process.Run(
        [
            "e3-sandbox",
            "exec",
            "--specs-dir",
            "nospecs",
            "--plan",
            "noplan",
            sandbox_dir,
        ]
    )
    assert no_specs.status != 0
    assert "nospecs does not exist" in no_specs.out


def test_sandbox_exec_api_version(git_specs_dir):
    """Test sandbox api version check."""
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)
    with open(os.path.join(sandbox_dir, "test.plan"), "w") as fd:
        fd.write("anod_build('dummyspec')\n")

    specs_source_dir = os.path.join(os.path.dirname(__file__), "specs")
    local_spec_dir = os.path.join(root_dir, "specs")

    e3.fs.sync_tree(specs_source_dir, local_spec_dir)
    create_prolog(local_spec_dir, "")


def test_sandbox_action_errors(git_specs_dir):
    """Test sandbox action error reporting."""
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)
    with open(os.path.join(sandbox_dir, "test.plan"), "w") as fd:
        fd.write("anod_build('builderror')\n")

    specs_source_dir = os.path.join(os.path.dirname(__file__), "specs")
    local_spec_dir = os.path.join(root_dir, "specs")

    e3.fs.sync_tree(specs_source_dir, local_spec_dir)
    create_prolog(local_spec_dir)

    # Test with local specs
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "exec",
            "--specs-dir",
            os.path.join(root_dir, "specs"),
            "--plan",
            os.path.join(sandbox_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    assert "'doesnotexist' is not defined" in p.out
    assert "builderror build fails" in p.out


def test_sandbox_exec_missing_plan_file(git_specs_dir):
    """Test sandbox exec exception.

    - Check if sandbox exec raises exception if plan file is missing
    """
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)

    # Plan file is missing
    no_plan = e3.os.process.Run(
        [
            "e3-sandbox",
            "exec",
            "--spec-git-url",
            git_specs_dir,
            "--plan",
            "noplan",
            "--create-sandbox",
            sandbox_dir,
        ]
    )
    assert no_plan.status != 0
    assert "SandBoxExec.run: plan file noplan does not exist" in no_plan.out


def test_sandbox_exec_success(git_specs_dir):
    """Test if sandbox exec works with local specs and a git repo."""
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    platform = e3.platform.Platform.get().platform

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)
    with open(os.path.join(sandbox_dir, "test.plan"), "w") as fd:
        fd.write("anod_build('dummyspec')\n")

    specs_source_dir = os.path.join(os.path.dirname(__file__), "specs")
    local_spec_dir = os.path.join(root_dir, "specs")

    e3.fs.sync_tree(specs_source_dir, local_spec_dir)
    create_prolog(local_spec_dir)

    e3.anod.helper.text_replace(
        os.path.join(local_spec_dir, "conf.yaml"),
        [(b"GITURL", git_specs_dir.encode("utf-8"))],
    )

    # Test with local specs
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "exec",
            "--specs-dir",
            local_spec_dir,
            "--plan",
            os.path.join(sandbox_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    assert "build dummyspec for %s" % platform in p.out
    assert "result: OK" in p.out

    with open(os.path.join(sandbox_dir, "test.plan"), "w") as fd:
        fd.write("anod_build('dummyspec')\n")
        fd.write("anod_test('dummyspec')\n")

    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "exec",
            "--specs-dir",
            local_spec_dir,
            "--plan",
            os.path.join(sandbox_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    assert "build dummyspec for %s" % platform in p.out
    assert "test dummyspec for %s" % platform in p.out
    assert "I am building" in p.out, p.out
    assert "I am testing" in p.out, p.out
    assert p.status == 0

    # Test with git module
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "exec",
            "--spec-git-url",
            git_specs_dir,
            "--plan",
            os.path.join(sandbox_dir, "test.plan"),
            "--dry-run",
            "--create-sandbox",
            sandbox_dir,
        ]
    )
    assert "build dummyspec for %s" % platform in p.out
    assert "test dummyspec for %s" % platform in p.out


def test_sandbox_source_auto_ignore(git_specs_dir):
    """Test if the new installation do not remove other sources."""
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)
    with open(os.path.join(sandbox_dir, "test.plan"), "w") as fd:
        fd.write("anod_build('autoignore')\n")

    specs_source_dir = os.path.join(os.path.dirname(__file__), "specs")
    local_spec_dir = os.path.join(root_dir, "specs")

    e3.fs.sync_tree(specs_source_dir, local_spec_dir)
    create_prolog(local_spec_dir)

    e3.anod.helper.text_replace(
        os.path.join(local_spec_dir, "conf.yaml"),
        [(b"GITURL", git_specs_dir.encode("utf-8"))],
    )

    # Test with local specs
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "exec",
            "--specs-dir",
            local_spec_dir,
            "--plan",
            os.path.join(sandbox_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    assert "found 2" in p.out
    assert "found 1" in p.out
    for subdir in ("3", "3/2", "3/2/1"):
        assert f"found .anod in {subdir}" in p.out
    assert "result: OK" in p.out


def test_sandbox_directory(git_specs_dir):
    """Test if the Source destination exists before prepare_src."""
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)
    with open(os.path.join(sandbox_dir, "test.plan"), "w") as fd:
        fd.write("anod_build('checkdirectory')\n")

    specs_source_dir = os.path.join(os.path.dirname(__file__), "specs")
    local_spec_dir = os.path.join(root_dir, "specs")

    e3.fs.sync_tree(specs_source_dir, local_spec_dir)
    create_prolog(local_spec_dir)

    e3.anod.helper.text_replace(
        os.path.join(local_spec_dir, "conf.yaml"),
        [(b"GITURL", git_specs_dir.encode("utf-8"))],
    )

    # Test with local specs
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "exec",
            "--specs-dir",
            local_spec_dir,
            "--plan",
            os.path.join(sandbox_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    assert "found file in destination directory" in p.out


def test_anod_plan(git_specs_dir):
    """Test if sandbox exec works with local specs and a git repo."""
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    # create sandbox
    with_git = e3.os.process.Run(
        ["e3-sandbox", "create", "--spec-git-url", git_specs_dir, sandbox_dir],
        output=None,
    )
    assert with_git.status == 0

    # Test anod build
    platform = e3.platform.Platform.get().platform

    # Build action
    with open(os.path.join(root_dir, "dummy_build.plan"), "w") as fd:
        fd.write("anod_build('dummyspec')")

    command = [
        "e3-sandbox",
        "exec",
        "--plan",
        os.path.join(root_dir, "dummy_build.plan"),
        "--dry-run",
        sandbox_dir,
    ]
    p = e3.os.process.Run(command)
    assert p.status == 0
    assert "build dummyspec for %s" % platform in p.out

    # Install action
    with open(os.path.join(root_dir, "dummy_build.plan"), "w") as fd:
        fd.write("anod_install('dummyspec')")
    p = e3.os.process.Run(command)
    assert p.status == 0
    actions = [
        "download binary of %s.dummyspec" % platform,
        "install dummyspec for %s" % platform,
    ]
    for action in actions:
        assert action in p.out

    # Test action
    with open(os.path.join(root_dir, "dummy_build.plan"), "w") as fd:
        fd.write("anod_test('dummyspec')")
    p = e3.os.process.Run(command)
    assert p.status == 0, p.out
    actions = [
        "checkout dummy-github",
        "create source dummy-src",
        "get source dummy-src",
        "install source dummy-src",
        "test dummyspec for %s" % platform,
    ]
    for action in actions:
        assert action in p.out

    # Test with download source resolver
    command = [
        "e3-sandbox",
        "exec",
        "--plan",
        os.path.join(root_dir, "dummy_build.plan"),
        "--dry-run",
        "--resolver",
        "always_download_source_resolver",
        sandbox_dir,
    ]
    p = e3.os.process.Run(command)
    assert p.status == 0
    actions = [
        "download source dummy-src",
        "get source dummy-src",
        "install source dummy-src",
        "test dummyspec for %s" % platform,
    ]
    for action in actions:
        assert action in p.out


def test_failure_status(git_specs_dir):
    """This test will run sandbox exec with an expected error.

    The error should be propagated throught the DAG and no action
    should be executed once we have the fail
    """
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    with open(os.path.join(root_dir, "test.plan"), "w") as fd:
        fd.write("anod_test('dummyspec')\n")

    specs_source_dir = os.path.join(os.path.dirname(__file__), "specs")
    local_spec_dir = os.path.join(root_dir, "specs")

    e3.fs.sync_tree(specs_source_dir, local_spec_dir)
    create_prolog(local_spec_dir)

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "-v",
            "exec",
            "--specs-dir",
            local_spec_dir,
            "--plan",
            os.path.join(root_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    # the dag for this plan has 6 actions and thus we need to have
    # 6 failure status (status=failure)
    assert p.out.count("status=failure") == 7, p.out
    assert "GITURL" in p.out, p.out

    # Try with an unsupported VCS
    e3.anod.helper.text_replace(
        os.path.join(local_spec_dir, "conf.yaml"),
        [(b"vcs: git", b"vcs: unsupported-vcs")],
    )
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "-v",
            "exec",
            "--specs-dir",
            local_spec_dir,
            "--plan",
            os.path.join(root_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    assert "unsupported-vcs vcs type not supported" in p.out, p.out
    assert p.out.count("status=failure") == 7, p.out

    # Also check with a missing repo
    e3.anod.helper.text_replace(
        os.path.join(local_spec_dir, "conf.yaml"), [(b"dummy-github", b"another_repo")]
    )
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "-v",
            "exec",
            "--specs-dir",
            local_spec_dir,
            "--plan",
            os.path.join(root_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    assert "add_spec: unknown repository dummy-github" in p.out, p.out
    assert p.status == 1


def test_sandbox_user_yaml(git_specs_dir):
    """Verify that user.yaml specs_dir is taken into account."""
    root_dir = os.getcwd()
    sandbox_dir = os.path.join(root_dir, "sbx")

    e3.os.process.Run(["e3-sandbox", "create", sandbox_dir], output=None)
    with open(os.path.join(sandbox_dir, "test.plan"), "w") as fd:
        fd.write("anod_build('dummyspec')\n")

    specs_source_dir = os.path.join(os.path.dirname(__file__), "specs")
    local_spec_dir = os.path.join(root_dir, "myspecs")

    with open(os.path.join(sandbox_dir, "user.yaml"), "w") as fd:
        yaml.dump({"specs_dir": local_spec_dir}, fd)

    e3.fs.sync_tree(specs_source_dir, local_spec_dir)
    create_prolog(local_spec_dir)
    # Test with local specs
    p = e3.os.process.Run(
        [
            "e3-sandbox",
            "-v",
            "exec",
            "--plan",
            os.path.join(sandbox_dir, "test.plan"),
            sandbox_dir,
        ]
    )
    assert "build dummyspec" in p.out
    assert f"using alternate specs dir {local_spec_dir}" in p.out

    sbx = e3.anod.sandbox.SandBox()
    sbx.root_dir = sandbox_dir

    assert sbx.is_alternate_specs_dir

    if sys.platform == "win32":
        assert sbx.specs_dir.lower() == local_spec_dir.lower()
    else:
        assert sbx.specs_dir == local_spec_dir



from __future__ import annotations

import stevedore


def load_store(name, configuration, cache):
    plugin = stevedore.DriverManager("e3.store", name)

    return plugin.driver(configuration, cache)



"""High-Level interface to Git repository.

Example::


    g = GitRepository(working_tree='/tmp/e3-core')
    g.init()
    g.update('ssh://git.adacore.com/anod', refspec='master', force=True)
    with open('/tmp/e3-core-log', 'w') as fd:
        g.write_log(fd, max_count=10)
    with open('/tmp/e3-core-log') as fd:
        authors = []
        for commit in g.parse_log(fd, max_diff_size=1024):
            authors.append(commit['email'])
"""


from __future__ import annotations

import itertools
import os
import sys
import tempfile
from contextlib import closing
from typing import TYPE_CHECKING

import e3.fs
import e3.log
import e3.os.fs
import e3.os.process
from e3.os.process import PIPE
from e3.text import bytes_as_str
from e3.vcs import VCSError

if TYPE_CHECKING:
    from typing import (
        Any,
        Dict,
        Final,
        Iterator,
        IO,
        List,
        Literal,
        Optional,
        TextIO,
        Union,
    )
    from e3.os.process import Run, STDOUT_VALUE, DEVNULL_VALUE, PIPE_VALUE

    Git_Cmd = List[Optional[str]]
    GIT_LOG_STREAM_VALUE = Literal[-4]

# Special value to direct outputs to the git log stream
# STDOUT, PIPE, and DEVNULL values are between -1 and -3
GIT_LOG_STREAM: GIT_LOG_STREAM_VALUE = -4

logger = e3.log.getLogger("vcs.git")

HEAD: Final = "HEAD"
FETCH_HEAD: Final = "FETCH_HEAD"


# Implementation note: some git commands can produce a big amount of data (e.g.
# git diff or git log). We always redirect git command result to a file then
# parse it, limiting the size of data read to avoid crashing our program.


class GitError(VCSError):
    pass


class GitRepository:
    """Interface to a Git Repository.

    :cvar git: path to the git binary
    :cvar log_stream: stream where the log commands will be redirected
        (default is stdout)
    :ivar working_tree: path to the git working tree
    """

    git: Optional[str] = None
    log_stream: Union[TextIO, IO[str]] = sys.stdout

    def __init__(self, working_tree: str):
        """Initialize a GitRepository object.

        :param working_tree: working tree of the GitRepository
        """
        self.working_tree = working_tree

    @classmethod
    def create(cls, repo_path: str, initial_content_path: Optional[str] = None) -> str:
        """Create a local Git repository.

        :param repo_path: a local directory where to create the repository
        :param initial_content_path: directory containing the initial content
            of the repository. If set to None an empty repository is created.
        :return: the URL of the newly created repository
        """
        repo_path = os.path.abspath(repo_path)
        repo = cls(repo_path)
        repo.init()
        if initial_content_path is not None:
            e3.fs.sync_tree(initial_content_path, repo_path, ignore=[".git"])
            repo.git_cmd(["add", "-A"])
            repo.git_cmd(["config", "user.email", "e3-core@example.net"])
            repo.git_cmd(["config", "user.name", "e3 core"])
            repo.git_cmd(["commit", "-m", "initial content"])
        return repo_path

    def git_cmd(
        self,
        cmd: Git_Cmd,
        output: Union[
            STDOUT_VALUE, DEVNULL_VALUE, PIPE_VALUE, GIT_LOG_STREAM_VALUE, str, IO, None
        ] = GIT_LOG_STREAM,
        **kwargs: Any
    ) -> Run:
        """Run a git command.

        :param cmd: the command line as a list of string, all None entries will
            be discarded
        :param output: see e3.os.process.Run, by default it is the
            ``log_stream`` class attribute.
        """
        if self.__class__.git is None:
            git_binary = e3.os.process.which("git", default=None)
            if git_binary is None:  # defensive code
                raise GitError("cannot find git", "git_cmd")
            self.__class__.git = e3.os.fs.unixpath(git_binary)

        if output == GIT_LOG_STREAM:
            output = self.log_stream

        p_cmd = [arg for arg in cmd if arg is not None]
        p_cmd.insert(0, self.__class__.git)

        p = e3.os.process.Run(p_cmd, cwd=self.working_tree, output=output, **kwargs)
        if p.status != 0:
            raise GitError(
                "{} failed (exit status: {})".format(
                    e3.os.process.command_line_image(p_cmd), p.status
                ),
                origin="git_cmd",
                process=p,
            )
        return p

    def init(self, url: Optional[str] = None, remote: Optional[str] = "origin") -> None:
        """Initialize a new Git repository and configure the remote.

        :param url: url of the remote repository, if None create a local git
            repository
        :param remote: name of the remote to create
        :raise: GitError
        """
        e3.fs.mkdir(self.working_tree)
        self.git_cmd(["init", "-q"])

        # Git version 1.8.3.1 might crash when calling "git stash" when
        # .git/logs/refs is not created. Recent versions of git do not
        # exhibit this problem
        e3.fs.mkdir(os.path.join(self.working_tree, ".git", "logs", "refs"))

        if url is not None:
            self.git_cmd(["remote", "add", remote, url])

    def checkout(self, branch: str, force: bool = False) -> None:
        """Checkout a given refspec.

        :param branch: name of the branch to checkout
        :param force: throw away local changes if needed
        :raise: GitError
        """
        cmd = ["checkout", "-q", "-f" if force else None, branch]
        self.git_cmd(cmd)

    def describe(self, commit: str = HEAD) -> str:
        """Get a human friendly revision for the given refspec.

        :param commit: commit object to describe.
        :return: the most recent tag name with the number of additional commits
            on top of the tagged object and the abbreviated object name of the
            most recent commit (see `git help describe`).
        :raise: GitError
        """
        p = self.git_cmd(["describe", "--always", commit], output=PIPE)
        return p.out.strip()

    def write_local_diff(self, stream: IO[str]) -> None:
        """Write local changes in the working tree in stream.

        :param stream: an open file descriptor
        :raise: GitError
        """
        cmd: Git_Cmd = ["--no-pager", "diff"]
        self.git_cmd(cmd, output=stream, error=self.log_stream)

    def write_diff(self, stream: IO[bytes], commit: str) -> None:
        """Write commit diff in stream.

        :param commit: revision naming a commit object, e.g. sha1 or symbolic
            refname
        :param stream: an open file descriptor
        :raise: GitError
        """
        cmd: Git_Cmd = [
            "--no-pager",
            "diff-tree",
            "--cc",
            "--no-commit-id",
            "--root",
            commit,
        ]
        self.git_cmd(cmd, output=stream, error=self.log_stream)

    def fetch(self, url: str, refspec: Optional[str] = None) -> None:
        """Fetch remote changes.

        :param url: url of the remote repository
        :param refspec: specifies which refs to fetch and which local refs to
            update.
        :raise: GitError
        """
        self.git_cmd(["fetch", url, refspec])

    def update(self, url: str, refspec: str, force: bool = False) -> None:
        """Fetch remote changes and checkout FETCH_HEAD.

        :param url: url of the remote repository
        :param refspec: specifies which refs to fetch and which local refs to
            update.
        :param force: throw away local changes if needed
        :raise: GitError
        """
        self.fetch(url, refspec)
        self.checkout("FETCH_HEAD", force=force)

    def fetch_gerrit_notes(self, url: str) -> None:
        """Fetch notes generated by Gerrit in `refs/notes/review`.

        :param url: url of the remote repository
        """
        self.fetch(url, "refs/notes/review:refs/notes/review")

    def write_log(
        self,
        stream: IO[str],
        max_count: int = 50,
        rev_range: Optional[str] = None,
        with_gerrit_notes: bool = False,
    ) -> None:
        """Write formatted log to a stream.

        :param stream: an open stream where to write the log content
        :param max_count: max number of commit to display
        :param rev_range: git revision range, see ``git log -h`` for details
        :param with_gerrit_notes: if True also fetch Gerrit notes containing
            review data such as Submitted-at, Submitted-by.
        :raise: GitError
        """
        # Format:
        #   %H: commit hash
        #   %aE: author email respecting .mailmap
        #   %ci: committer date, ISO 8601-like format (don't use %cI)
        #   %n: new line
        #   %B: raw body (unwrapped subject and body)
        #   %N: commit notes
        cmd: Git_Cmd = [
            "log",
            "--format=format:%H%n%aE%n%ci%n"
            + ("%N%n" if with_gerrit_notes else "")
            + "%n%B",
            "--log-size",
            "--max-count=%d" % max_count if max_count else None,
            "--show-notes=review" if with_gerrit_notes else None,
            rev_range,
        ]
        self.git_cmd(cmd, output=stream, error=None)

    def parse_log(
        self, stream: IO[str], max_diff_size: int = 0
    ) -> Iterator[Dict[str, Union[str, dict]]]:
        """Parse a log stream generated with `write_log`.

        :param stream: stream of text to read
        :param max_diff_size: max size of a diff, if <= 0 diff are ignored
        :return: a generator returning commit information (directories with
            the following keys: sha, email, date, notes, message, diff). Note
            that the key diff is only set when max_diff_size is bigger than 0.
            The notes value is a dictionary built from the 'key:value' found in
            Gerrit notes.
        """

        def to_commit(object_content: str) -> dict:
            """Return commit information."""
            headers, body = object_content.split("\n\n", 1)

            # Retrieve sha, email, date, and (optionally) notes if some notes
            # are attached to the commit
            result = dict(
                itertools.zip_longest(
                    ("sha", "email", "date", "notes"),
                    headers.replace("\r", "").split("\n", 3),
                )
            )

            # replace notes "key: value" lines by a dictionary
            if result["notes"]:
                try:
                    result["notes"] = dict(
                        notes_line.split(": ", 1)
                        for notes_line in result["notes"].splitlines()
                    )
                except ValueError:
                    # Notes format invalid, discard it
                    result["notes"] = None

            result["message"] = body

            if max_diff_size > 0:

                tempfile_name = None
                try:
                    with closing(
                        tempfile.NamedTemporaryFile(mode="wb", delete=False)
                    ) as diff_fd:
                        tempfile_name = diff_fd.name
                        self.write_diff(diff_fd, result["sha"])

                        # compute file size
                        diff_fd.seek(0, 2)
                        diff_size = diff_fd.tell()
                        e3.log.debug("diff size for %s: %d", result["sha"], diff_size)

                    with open(tempfile_name, "rb") as f:
                        content = f.read(max_diff_size)

                        # Diff content is not always in utf-8 format thus use
                        # a safe function to decode it.
                        result["diff"] = bytes_as_str(content)

                    if diff_size > max_diff_size:
                        result["diff"] += "\n... diff too long ...\n"
                finally:
                    if tempfile_name is not None:
                        e3.fs.rm(tempfile_name)

            return result

        size_to_read = 0
        while True:
            line = stream.readline()
            if not line.strip():
                # Strip empty line separating two commits
                line = stream.readline()
            if line.startswith("log size "):
                size_to_read = int(line.rsplit(None, 1)[1])
                # Get commit info
            if size_to_read <= 0:
                return
            yield to_commit(stream.read(size_to_read))
            size_to_read = 0

    def rev_parse(self, refspec: str = HEAD) -> str:
        """Get the sha associated to a given refspec.

        :param refspec: refspec.
        :raise: GitError
        """
        p = self.git_cmd(["rev-parse", "--revs-only", refspec], output=PIPE, error=PIPE)
        return p.out.strip()
